import unittest

import numpy as np
from numpy.testing import assert_almost_equal
import openmdao.api as om
from dymos.utils.testing_utils import assert_check_partials

from dymos.transcriptions.pseudospectral.components import StateInterpComp
from dymos.transcriptions.grid_data import GridData
from dymos.utils.lgr import lgr

# Modify class so we can run it standalone.
import dymos as dm
from dymos.utils.misc import CompWrapperConfig
StateInterpComp = CompWrapperConfig(StateInterpComp)

SHOW_PLOTS = False

if SHOW_PLOTS:
    import matplotlib.pyplot as plt


# Test 1:  Let x = t**2, f = 2*t
def x(t):
    return t ** 2


def f_x(t):
    return 2 * t


# Test 1:  Let v = t**3-10*t**2, f = 3*t**2 - 20*t
def v(t):
    return t ** 3 - 10 * t ** 2


def f_v(t):
    return 3 * t ** 2 - 20 * t


class TestStateInterpComp(unittest.TestCase):

    def setUp(self):
        dm.options['include_check_partials'] = True

    def tearDown(self):
        dm.options['include_check_partials'] = False

    def test_state_interp_comp_lobatto(self):

        segends = np.array([0.0, 3.0, 10.0])

        gd = GridData(num_segments=2,
                      transcription_order=3,
                      segment_ends=segends,
                      transcription='gauss-lobatto')

        p = om.Problem(model=om.Group())

        states = {'x': {'units': 'm', 'shape': (1,)},
                  'v': {'units': 'm/s', 'shape': (1,)}}

        X_ivc = om.IndepVarComp()
        p.model.add_subsystem('X_ivc', X_ivc, promotes=['state_disc:x', 'state_disc:v'])

        X_ivc.add_output('state_disc:x', val=np.zeros(gd.subset_num_nodes['state_disc']),
                         units='m')

        X_ivc.add_output('state_disc:v', val=np.zeros(gd.subset_num_nodes['state_disc']),
                         units='m/s')

        F_ivc = om.IndepVarComp()
        p.model.add_subsystem('F_ivc', F_ivc, promotes=['staterate_disc:x', 'staterate_disc:v'])

        F_ivc.add_output('staterate_disc:x',
                         val=np.zeros(gd.subset_num_nodes['state_disc']),
                         units='m/s')

        F_ivc.add_output('staterate_disc:v',
                         val=np.zeros(gd.subset_num_nodes['state_disc']),
                         units='m/s**2')

        dt_dtau_ivc = om.IndepVarComp()
        p.model.add_subsystem('dt_dstau_ivc', dt_dtau_ivc, promotes=['dt_dstau'])

        dt_dtau_ivc.add_output('dt_dstau', val=0.0*np.zeros(gd.subset_num_nodes['col']), units='s')

        p.model.add_subsystem('state_interp_comp',
                              subsys=StateInterpComp(transcription='gauss-lobatto',
                                                     grid_data=gd,
                                                     state_options=states,
                                                     time_units='s'))

        p.model.connect('state_disc:x', 'state_interp_comp.state_disc:x')
        p.model.connect('state_disc:v', 'state_interp_comp.state_disc:v')
        p.model.connect('staterate_disc:x', 'state_interp_comp.staterate_disc:x')
        p.model.connect('staterate_disc:v', 'state_interp_comp.staterate_disc:v')
        p.model.connect('dt_dstau', 'state_interp_comp.dt_dstau')

        p.setup(force_alloc_complex=True)

        segends_disc = segends[np.array((0, 1, 1, 2), dtype=int)]

        p['state_disc:x'] = [x(t) for t in segends_disc]
        p['staterate_disc:x'] = [f_x(t) for t in segends_disc]

        p['state_disc:v'] = [v(t) for t in segends_disc]
        p['staterate_disc:v'] = [f_v(t) for t in segends_disc]

        p['dt_dstau'] = (segends[1:] - segends[:-1]) / 2.0

        p.run_model()

        t_disc = segends_disc
        t_col = (segends[1:] + segends[:-1]) / 2.0

        if SHOW_PLOTS:  # pragma: no cover
            f, ax = plt.subplots(2, 1)

            t = np.linspace(0, 10, 100)
            x1 = x(t)
            xdot1 = f_x(t)

            x2 = v(t)
            xdot2 = f_v(t)

            ax[0].plot(t, x1, 'b-', label='$x$')
            ax[0].plot(t, xdot1, 'b--', label=r'$\dot{x}$')
            ax[0].plot(t_disc, p['state_disc:x'], 'bo', label='$X_d:x$')
            ax[0].plot(t_col, p['state_interp_comp.state_col:x'], 'bv', label='$X_c:x$')
            ax[0].plot(t_col, p['state_interp_comp.staterate_col:x'], marker='v', color='None',
                       mec='b', label='$Xdot_c:x$')

            ax[1].plot(t, x2, 'r-', label='$v$')
            ax[1].plot(t, xdot2, 'r--', label=r'$\dot{v}$')
            ax[1].plot(t_disc, p['state_disc:v'], 'ro', label='$X_d:v$')
            ax[1].plot(t_col, p['state_interp_comp.state_col:v'], 'rv', label='$X_c:v$')
            ax[1].plot(t_col, p['state_interp_comp.staterate_col:v'], marker='v', color='None',
                       mec='r', label='$Xdot_c:v$')

            ax[0].legend(loc='upper left', ncol=3)
            ax[1].legend(loc='upper left', ncol=3)

            plt.show()

        # Test 1
        assert_almost_equal(
            p['state_interp_comp.state_col:x'][:, 0], x(t_col))
        assert_almost_equal(
            p['state_interp_comp.staterate_col:x'][:, 0], f_x(t_col))

        # Test 2
        assert_almost_equal(
            p['state_interp_comp.state_col:v'][:, 0], v(t_col))
        assert_almost_equal(
            p['state_interp_comp.staterate_col:v'][:, 0], f_v(t_col))

        cpd = p.check_partials(compact_print=True, method='cs')
        assert_check_partials(cpd, atol=5.0E-5)

    def test_state_interp_comp_lobatto_vectorized(self):

        segends = np.array([0.0, 3.0, 10.0])

        gd = GridData(num_segments=2,
                      transcription_order=3,
                      segment_ends=segends,
                      transcription='gauss-lobatto')

        p = om.Problem(model=om.Group())

        states = {'pos': {'units': 'm', 'shape': (2,)}}

        X_ivc = om.IndepVarComp()
        p.model.add_subsystem('X_ivc', X_ivc, promotes=['state_disc:pos'])

        X_ivc.add_output('state_disc:pos',
                         val=np.zeros((gd.subset_num_nodes['state_disc'], 2)), units='m')

        F_ivc = om.IndepVarComp()
        p.model.add_subsystem('F_ivc', F_ivc, promotes=['staterate_disc:pos'])

        F_ivc.add_output('staterate_disc:pos',
                         val=np.zeros((gd.subset_num_nodes['state_disc'], 2)),
                         units='m/s')

        dt_dtau_ivc = om.IndepVarComp()
        p.model.add_subsystem('dt_dstau_ivc', dt_dtau_ivc, promotes=['dt_dstau'])

        dt_dtau_ivc.add_output('dt_dstau', val=0.0*np.zeros(gd.subset_num_nodes['col']), units='s')

        p.model.add_subsystem('state_interp_comp',
                              subsys=StateInterpComp(transcription='gauss-lobatto',
                                                     grid_data=gd,
                                                     state_options=states,
                                                     time_units='s'))

        p.model.connect('state_disc:pos', 'state_interp_comp.state_disc:pos')
        p.model.connect('staterate_disc:pos', 'state_interp_comp.staterate_disc:pos')
        p.model.connect('dt_dstau', 'state_interp_comp.dt_dstau')

        p.setup(force_alloc_complex=True)

        segends_disc = segends[np.array((0, 1, 1, 2), dtype=int)]

        p['state_disc:pos'][:, 0] = [x(t) for t in segends_disc]  # [0.0, 25.0, 25.0, 100.0]
        p['staterate_disc:pos'][:, 0] = [f_x(t) for t in segends_disc]

        p['state_disc:pos'][:, 1] = [v(t) for t in segends_disc]
        p['staterate_disc:pos'][:, 1] = [f_v(t) for t in segends_disc]

        p['dt_dstau'] = (segends[1:] - segends[:-1]) / 2.0

        p.run_model()

        t_disc = segends_disc
        t_col = (segends[1:] + segends[:-1]) / 2.0

        if SHOW_PLOTS:  # pragma: no cover
            f, ax = plt.subplots(2, 1)

            print(t_disc)
            print(t_col)
            print(p['dt_dstau'])
            print(p['state_disc:pos'][:, 0])
            print(p['staterate_disc:pos'][:, 0])
            print(p['state_disc:pos'][:, 0])
            print(p['staterate_disc:pos'][:, 1])

            t = np.linspace(0, 10, 100)
            x1 = x(t)
            xdot1 = f_x(t)

            x2 = v(t)
            xdot2 = f_v(t)

            ax[0].plot(t, x1, 'b-', label='$x$')
            ax[0].plot(t, xdot1, 'b--', label='r$\dot{x}$')
            ax[0].plot(t_disc, p['state_disc:pos'][:, 0], 'bo', label='$X_d:pos$')
            ax[0].plot(t_col, p['state_interp_comp.state_col:pos'][:, 0], 'bv', label='$X_c:pos$')
            ax[0].plot(t_col, p['state_interp_comp.staterate_col:pos'][:, 0], marker='v',
                       color='None', mec='b', label='$Xdot_c:pos$')

            ax[1].plot(t, x2, 'r-', label='$v$')
            ax[1].plot(t, xdot2, 'r--', label='r$\dot{v}$')
            ax[1].plot(t_disc, p['state_disc:pos'][:, 1], 'ro', label='$X_d:vel$')
            ax[1].plot(t_col, p['state_interp_comp.state_col:pos'][:, 1], 'rv', label='$X_c:vel$')
            ax[1].plot(t_col, p['state_interp_comp.staterate_col:pos'][:, 1], marker='v',
                       color='None', mec='r', label='$Xdot_c:vel$')

            ax[0].legend(loc='upper left', ncol=3)
            ax[1].legend(loc='upper left', ncol=3)

            plt.show()

        # Test 1
        assert_almost_equal(
            p['state_interp_comp.state_col:pos'][:, 0], x(t_col))
        assert_almost_equal(
            p['state_interp_comp.staterate_col:pos'][:, 0], f_x(t_col))

        # Test 2
        assert_almost_equal(
            p['state_interp_comp.state_col:pos'][:, 1], v(t_col))
        assert_almost_equal(
            p['state_interp_comp.staterate_col:pos'][:, 1], f_v(t_col))

        cpd = p.check_partials(compact_print=True, method='cs')
        assert_check_partials(cpd, atol=5.0E-5)

    def test_state_interp_comp_lobatto_vectorized_different_orders(self):

        segends = np.array([0.0, 3.0, 10.0])

        gd = GridData(num_segments=2,
                      transcription_order=[3, 5],
                      segment_ends=segends,
                      transcription='gauss-lobatto')

        p = om.Problem(model=om.Group())

        states = {'pos': {'units': 'm', 'shape': (2,)}}

        X_ivc = om.IndepVarComp()
        p.model.add_subsystem('X_ivc', X_ivc, promotes=['state_disc:pos'])

        X_ivc.add_output('state_disc:pos',
                         val=np.zeros((gd.subset_num_nodes['state_disc'], 2)), units='m')

        F_ivc = om.IndepVarComp()
        p.model.add_subsystem('F_ivc', F_ivc, promotes=['staterate_disc:pos'])

        F_ivc.add_output('staterate_disc:pos',
                         val=np.zeros((gd.subset_num_nodes['state_disc'], 2)),
                         units='m/s')

        dt_dtau_ivc = om.IndepVarComp()
        p.model.add_subsystem('dt_dstau_ivc', dt_dtau_ivc, promotes=['dt_dstau'])

        dt_dtau_ivc.add_output('dt_dstau', val=0.0*np.zeros(gd.subset_num_nodes['col']), units='s')

        p.model.add_subsystem('state_interp_comp',
                              subsys=StateInterpComp(transcription='gauss-lobatto',
                                                     grid_data=gd,
                                                     state_options=states,
                                                     time_units='s'))

        p.model.connect('state_disc:pos', 'state_interp_comp.state_disc:pos')
        p.model.connect('staterate_disc:pos', 'state_interp_comp.staterate_disc:pos')
        p.model.connect('dt_dstau', 'state_interp_comp.dt_dstau')

        p.setup(force_alloc_complex=True)

        segends_disc = np.array((0, 3, 3, 6.5, 10))

        p['state_disc:pos'][:, 0] = [x(t) for t in segends_disc]  # [0.0, 25.0, 25.0, 100.0]
        p['staterate_disc:pos'][:, 0] = [f_x(t) for t in segends_disc]

        p['state_disc:pos'][:, 1] = [v(t) for t in segends_disc]
        p['staterate_disc:pos'][:, 1] = [f_v(t) for t in segends_disc]

        p['dt_dstau'] = [3.0/2., 7.0/2, 7.0/2]

        p.run_model()

        cpd = p.check_partials(compact_print=True, method='cs')
        assert_check_partials(cpd, atol=5.0E-5)

    def test_state_interp_comp_radau(self):

        gd = GridData(num_segments=1,
                      transcription_order=3,
                      segment_ends=np.array([0, 10]),
                      transcription='radau-ps')

        p = om.Problem(model=om.Group())

        states = {'x': {'units': 'm', 'shape': (1,)},
                  'v': {'units': 'm/s', 'shape': (1,)}}

        X_ivc = om.IndepVarComp()
        p.model.add_subsystem('X_ivc', X_ivc, promotes=['state_disc:x', 'state_disc:v'])

        X_ivc.add_output('state_disc:x', val=np.zeros(gd.subset_num_nodes['state_disc']),
                         units='m')

        X_ivc.add_output('state_disc:v', val=np.zeros(gd.subset_num_nodes['state_disc']),
                         units='m/s')

        dt_dtau_ivc = om.IndepVarComp()
        dt_dtau_ivc.add_output('dt_dstau', val=0.0*np.zeros(gd.subset_num_nodes['col']), units='s')

        p.model.add_subsystem('dt_dstau_ivc', dt_dtau_ivc, promotes=['dt_dstau'])

        p.model.add_subsystem('state_interp_comp',
                              subsys=StateInterpComp(transcription='radau-ps',
                                                     grid_data=gd,
                                                     state_options=states,
                                                     time_units='s'))

        p.model.connect('state_disc:x', 'state_interp_comp.state_disc:x')
        p.model.connect('state_disc:v', 'state_interp_comp.state_disc:v')
        p.model.connect('dt_dstau', 'state_interp_comp.dt_dstau')

        p.setup(force_alloc_complex=True)

        lgr_nodes, lgr_weights = lgr(3, include_endpoint=True)
        t_disc = (lgr_nodes + 1.0) * 5.0
        t_col = t_disc[:-1]

        # Test 1:  Let x = t**2, f = 2*t
        p['state_disc:x'] = t_disc**2

        # Test 1:  Let v = t**3-10*t**2, f = 3*t**2 - 20*t
        p['state_disc:v'] = t_disc**3-10*t_disc**2

        p['dt_dstau'] = 10/2.0

        p.run_model()

        if SHOW_PLOTS:  # pragma: no cover
            f, ax = plt.subplots(2, 1)

            t_disc = np.array([0, 5, 10])
            t_col = np.array([2.5, 7.5])

            t = np.linspace(0, 10, 100)
            x1 = t**2
            xdot1 = 2*t

            x2 = t**3 - 10*t**2
            xdot2 = 3*t**2 - 20*t

            ax[0].plot(t, x1, 'b-', label='$x$')
            ax[0].plot(t, xdot1, 'b--', label='$\dot{x}$')
            ax[0].plot(t_disc, p['state_disc:x'], 'bo', label='$X_d:x$')
            ax[0].plot(t_col, p['state_interp_comp.state_col:x'], 'bv', label='$X_c:x$')
            ax[0].plot(t_col, p['state_interp_comp.staterate_col:x'], marker='v', color='None',
                       mec='b', label='$Xdot_c:x$')

            ax[1].plot(t, x2, 'r-', label='$v$')
            ax[1].plot(t, xdot2, 'r--', label='$\dot{v}$')
            ax[1].plot(t_disc, p['state_disc:v'], 'ro', label='$X_d:v$')
            ax[1].plot(t_col, p['state_interp_comp.state_col:v'], 'rv', label='$X_c:v$')
            ax[1].plot(t_col, p['state_interp_comp.staterate_col:v'], marker='v', color='None',
                       mec='r', label='$Xdot_c:v$')

            ax[0].legend(loc='upper left', ncol=3)
            ax[1].legend(loc='upper left', ncol=3)

            plt.show()

        # Test 1
        assert_almost_equal(
            p['state_interp_comp.staterate_col:x'][:, 0], 2*t_col)

        # Test 2
        assert_almost_equal(
            p['state_interp_comp.staterate_col:v'][:, 0], 3*t_col**2 - 20*t_col)

        cpd = p.check_partials(compact_print=True, method='cs')
        assert_check_partials(cpd, atol=1.0E-5)


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



import numpy as np


def hermite_matrices(x_given, x_eval):
    """
    Return matrices for a Hermite polynomial at the given nodes.

    This includes interpolation matrices (A_i and B_i) and differentiation matrices (A_d and B_d).

    Parameters
    ----------
    x_given : ndarray[:]
        Vector of given nodes in the polynomial.
    x_eval : ndarray[:]
        Vector of nodes at which the polynomial is evaluated.

    Returns
    -------
    A_i : np.array
        An num_disc_nodes-1 x num_disc_nodes matrix used for the interpolation of state values
        at the interior LGL nodes.
    B_i : np.array
        An num_disc_nodes-1 x num_disc_nodes matrix used for the interpolation of state values
        at the interior LGL nodes.
    A_d : np.array
        An num_disc_nodes-1 x num_disc_nodes matrix used for the differentiation of state values
        at the collocation LGL nodes.
    B_d : np.array
        An num_disc_nodes-1 x num_disc_nodes matrix used for the differentiation of state values
        at the collocation LGL nodes.

    Notes
    -----
    .. math::
        x_i = \\left[ A_i\\right] x_c + \\frac{dt}{dtau} \\left[ B_i \\right] f_c
    """
    tau_disc = x_given
    """ Array of the discretization nodes. """

    num_disc_nodes = len(x_given)
    """ Number of discretization nodes per segment. """

    tau_col = x_eval
    """ Array of the collocation nodes. """

    num_col_nodes = len(x_eval)
    """ Number of collocation nodes per segment. """

    Ai = np.zeros([num_col_nodes, num_disc_nodes])
    Bi = np.zeros([num_col_nodes, num_disc_nodes])
    Ad = np.zeros([num_col_nodes, num_disc_nodes])
    Bd = np.zeros([num_col_nodes, num_disc_nodes])

    # The state interpolation matrices
    for i in range(num_col_nodes):
        ui, vi = heriwi(tau_col[i], tau_disc)
        Ai[i, :] = ui
        Bi[i, :] = vi

    # The differentiation matrices
    for i in range(num_col_nodes):
        ui, vi = heriwd(tau_col[i], tau_disc)
        Ad[i, :] = ui
        Bd[i, :] = vi

    return Ai, Bi, Ad, Bd


def heriwi(tau, taus):
    """
    Computes the weights for computing the values of polynomial with hermite interpolation.

    Based off of the ACM211 algorithm.

    Parameters
    ----------
    tau : float
        Value at which the Hermite polynomial weights are desired.
    taus : np.array
        Array of points at which the values and derivatives which
        define the Hermite polynomial are provided.

    Returns
    -------
    u : np.array
        Weights for function values.
    v : np.array
        Weights for derivative values.
    """
    n = len(taus)
    u = np.zeros(n)
    v = np.zeros(n)

    for j in range(n):
        prod = 1.0
        sum1 = 0.0
        for i in range(n):
            if i != j:
                prod *= ((tau - taus[i]) / (taus[j] - taus[i]))**2
                sum1 += 1. / (taus[j] - taus[i])
        u[j] = prod * ((taus[j] - tau) * 2.0 * sum1 + 1.0)
        v[j] = prod * (tau - taus[j])

    return u, v


def heriwd(tau, taus):
    """
    Computes the weights for computing the derivatives of a polynomial with hermite interpolation.

    Based off of the ACM211 algorithm

    Parameters
    ----------
    tau : float
        Value at which the Hermite polynomial weights are desired.
    taus : np.array
        Array of points at which the values and derivatives which
        define the Hermite polynomial are provided.

    Returns
    -------
    u : np.array
        Weights for function values.
    v : np.array
        Weights for derivative values.
    """
    n = len(taus)
    u = np.zeros(n)
    v = np.zeros(n)

    for j in range(n):
        prod = 1.0
        dprod = 0.0
        sum1 = 0.0
        for i in range(n):
            if i != j:
                xmxi = tau - taus[i]
                xjmxi = taus[j] - taus[i]
                dprod = dprod * (xmxi / xjmxi)**2 + 2.0 * prod * xmxi / xjmxi**2
                prod = prod * (xmxi / xjmxi)**2
                sum1 = sum1 + 1.0 / xjmxi
        xmxj = tau - taus[j]
        xjmx = taus[j] - tau
        u[j] = dprod * (xjmx * 2.0 * sum1 + 1.0) - prod * (2.0 * sum1)
        v[j] = dprod * xmxj + prod

    return u, v



import unittest

import numpy as np

import openmdao.api as om

from openmdao.utils.assert_utils import assert_near_equal, assert_check_partials
from openmdao.utils.testing_utils import use_tempdirs

from dymos.examples.finite_burn_orbit_raise.finite_burn_eom import FiniteBurnODE


@use_tempdirs
class TestFiniteBurnEOM(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        nn = 2

        p = cls.p = om.Problem(model=om.Group())

        ivc = p.model.add_subsystem('ivc', om.IndepVarComp(), promotes_outputs=['*'])

        p.model.add_subsystem('ode', subsys=FiniteBurnODE(num_nodes=nn), promotes_inputs=['*'],
                              promotes_outputs=['*'])

        ivc.add_output('r',
                       val=np.ones(nn),
                       desc='radius from center of attraction',
                       units='DU')

        ivc.add_output('theta',
                       val=np.zeros(nn),
                       desc='anomaly term',
                       units='rad')

        ivc.add_output('vr',
                       val=np.zeros(nn),
                       desc='local vertical velocity component',
                       units='DU/TU')

        ivc.add_output('vt',
                       val=np.zeros(nn),
                       desc='local horizontal velocity component',
                       units='DU/TU')

        ivc.add_output('accel',
                       val=np.zeros(nn),
                       desc='acceleration due to thrust',
                       units='DU/TU**2')

        ivc.add_output('u1',
                       val=np.zeros(nn),
                       desc='thrust angle above local horizontal',
                       units='rad')

        ivc.add_output('c',
                       val=np.zeros(nn),
                       desc='exhaust velocity',
                       units='DU/TU')

        p.setup()

        p['r'] = 2.0
        p['theta'] = 0.05
        p['vr'] = 0.2
        p['vt'] = 1.0
        p['accel'] = [0.1, 0.1]
        p['u1'] = [0.0, np.pi]
        p['c'] = 9.80665 * 2000

        p.run_model()

    def test_outputs(self):
        p = self.p
        assert_near_equal(p['r_dot'], p['vr'])
        assert_near_equal(p['theta_dot'], p['vt'] / p['r'])
        assert_near_equal(p['at_dot'], p['accel']**2 / p['c'])

    def test_partials(self):
        cpd = self.p.check_partials(compact_print=False)
        assert_check_partials(cpd, atol=1.0E-5, rtol=2.0)


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



import unittest

import numpy as np

import openmdao.api as om

from openmdao.utils.assert_utils import assert_near_equal
from openmdao.utils.testing_utils import use_tempdirs

from dymos.utils.testing_utils import assert_check_partials

from dymos.examples.aircraft_steady_flight.flight_equlibrium.steady_flight_equilibrium_group \
    import SteadyFlightEquilibriumGroup


@use_tempdirs
class TestFlightEquilibriumGroup(unittest.TestCase):

    def setUp(self):
        self.n = 10

        self.p = om.Problem(model=om.Group())

        ivc = self.p.model.add_subsystem('ivc', om.IndepVarComp(), promotes_outputs=['*'])

        ivc.add_output('alt', val=3000 * np.ones(self.n), units='m', desc='altitude above MSL')
        ivc.add_output('TAS', val=250.0 * np.ones(self.n), units='m/s', desc='true airspeed')
        # ivc.add_output('TAS_rate', val=np.zeros(self.n), units='m/s**2', desc='acceleration')
        ivc.add_output('gam', val=np.zeros(self.n), units='rad', desc='flight path angle')
        # ivc.add_output('gam_rate', val=np.zeros(self.n), units='rad/s',
        #                desc='flight path angle rate')
        ivc.add_output('S', val=427.8 * np.ones(self.n), units='m**2', desc='reference area')
        ivc.add_output('mach', val=0.8 * np.ones(self.n), units=None, desc='mach number')
        ivc.add_output('W_total', val=200000 * 9.80665 * np.ones(self.n), units='N',
                       desc='aircraft total weight')
        ivc.add_output('q', val=0.5 * 250.0**2 * np.ones(self.n), units='Pa',
                       desc='dynamic pressure')

        self.p.model.add_subsystem('flight_equilibrium',
                                   subsys=SteadyFlightEquilibriumGroup(num_nodes=self.n),
                                   promotes_inputs=['aero.*'],
                                   promotes_outputs=['aero.*'])

        self.p.model.connect('gam', 'flight_equilibrium.gam')
        self.p.model.connect('S', ('aero.S', 'flight_equilibrium.S'))
        self.p.model.connect('mach', 'aero.mach')
        self.p.model.connect('W_total', 'flight_equilibrium.W_total')
        self.p.model.connect('q', ('aero.q', 'flight_equilibrium.q'))

        self.p.setup(check=True, force_alloc_complex=True)

        self.p.run_model()

    def test_results(self):
        CL_eq = self.p['flight_equilibrium.CL_eq']
        CL = self.p['aero.CL']
        CM = self.p['aero.CM']

        assert_near_equal(CL_eq, CL, tolerance=1.0E-12)
        assert_near_equal(CM, np.zeros_like(CM), tolerance=1.0E-12)

    def test_partials(self):
        cpd = self.p.check_partials(out_stream=None, method='fd', step=1.0E-6)
        assert_check_partials(cpd, atol=5.0E-3, rtol=2.0)



import unittest

import openmdao.api as om
import openmdao.func_api as omf
from openmdao.utils.assert_utils import assert_near_equal
from openmdao.utils.testing_utils import use_tempdirs, require_pyoptsparse
from openmdao import __version__ as om_version
import dymos as dm
import numpy as np

try:
    import jax
except ImportError:
    jax = None


def runway_ode(rho, S, CD0, CL0, CL_max, alpha_max, h_w, AR, e, span, T, mu_r, m, v, h, alpha):
    g = 9.80665
    W = m * g
    v_stall = np.sqrt(2 * W / rho / S / CL_max)
    v_over_v_stall = v / v_stall

    CL = CL0 + (alpha / alpha_max) * (CL_max - CL0)
    K_nom = 1.0 / (np.pi * AR * e)
    b = span / 2.0
    fact = ((h + h_w) / b) ** 1.5
    K = K_nom * 33 * fact / (1.0 + 33 * fact)

    q = 0.5 * rho * v ** 2
    L = q * S * CL
    D = q * S * (CD0 + K * CL ** 2)

    # Compute the downward force on the landing gear
    calpha = np.cos(alpha)
    salpha = np.sin(alpha)

    # Runway normal force
    F_r = m * g - L * calpha - T * salpha

    # Compute the dynamics
    v_dot = (T * calpha - D - F_r * mu_r) / m
    r_dot = v

    return CL, q, L, D, K, F_r, v_dot, r_dot, W, v_stall, v_over_v_stall


def climb_ode(rho, S, CD0, CL0, CL_max, alpha_max, h_w, AR, e, span, T, m, v, h, alpha, gam):
    g = 9.80665
    W = m * g
    v_stall = np.sqrt(2 * W / rho / S / CL_max)
    v_over_v_stall = v / v_stall

    CL = CL0 + (alpha / alpha_max) * (CL_max - CL0)
    K_nom = 1.0 / (np.pi * AR * e)
    b = span / 2.0
    fact = ((h + h_w) / b) ** 1.5
    K = K_nom * 33 * fact / (1.0 + 33 * fact)

    q = 0.5 * rho * v ** 2
    L = q * S * CL
    D = q * S * (CD0 + K * CL ** 2)

    # Compute the downward force on the landing gear
    calpha = np.cos(alpha)
    salpha = np.sin(alpha)

    # Runway normal force
    F_r = m * g - L * calpha - T * salpha

    # Compute the dynamics
    cgam = np.cos(gam)
    sgam = np.sin(gam)
    v_dot = (T * calpha - D) / m - g * sgam
    gam_dot = (T * salpha + L) / (m * v) - (g / v) * cgam
    h_dot = v * sgam
    r_dot = v * cgam

    return CL, q, L, D, K, F_r, v_dot, r_dot, W, v_stall, v_over_v_stall, gam_dot, h_dot


def wrap_ode_func(num_nodes, flight_mode, grad_method='jax', jax_jit=True):
    """
    Returns the metadata from omf needed to create a new ExplciitFuncComp.
    """
    nn = num_nodes
    ode_func = runway_ode if flight_mode == 'runway' else climb_ode

    meta = (omf.wrap(ode_func)
            .add_input('rho', val=1.225, desc='atmospheric density at runway', units='kg/m**3')
            .add_input('S', val=124.7, desc='aerodynamic reference area', units='m**2')
            .add_input('CD0', val=0.03, desc='zero-lift drag coefficient', units=None)
            .add_input('CL0', val=0.5, desc='zero-alpha lift coefficient', units=None)
            .add_input('CL_max', val=2.0, desc='maximum lift coefficient for linear fit', units=None)
            .add_input('alpha_max', val=np.radians(10), desc='angle of attack at CL_max', units='rad')
            .add_input('h_w', val=1.0, desc='height of the wing above the CG', units='m')
            .add_input('AR', val=9.45, desc='wing aspect ratio', units=None)
            .add_input('e', val=0.801, desc='Oswald span efficiency factor', units=None)
            .add_input('span', val=35.7, desc='Wingspan', units='m')
            .add_input('T', val=1.0, desc='thrust', units='N')

            # Dynamic inputs (can assume a different value at every node)
            .add_input('m', shape=(nn,), desc='aircraft mass', units='kg')
            .add_input('v', shape=(nn,), desc='aircraft true airspeed', units='m/s')
            .add_input('h', shape=(nn,), desc='altitude', units='m')
            .add_input('alpha', shape=(nn,), desc='angle of attack', units='rad')

            # Outputs
            .add_output('CL', shape=(nn,), desc='lift coefficient', units=None)
            .add_output('q', shape=(nn,), desc='dynamic pressure', units='Pa')
            .add_output('L', shape=(nn,), desc='lift force', units='N')
            .add_output('D', shape=(nn,), desc='drag force', units='N')
            .add_output('K', val=np.ones(nn), desc='drag-due-to-lift factor', units=None)
            .add_output('F_r', shape=(nn,), desc='runway normal force', units='N')
            .add_output('v_dot', shape=(nn,), desc='rate of change of speed', units='m/s**2',
                        tags=['dymos.state_rate_source:v'])
            .add_output('r_dot', shape=(nn,), desc='rate of change of range', units='m/s',
                        tags=['dymos.state_rate_source:r'])
            .add_output('W', shape=(nn,), desc='aircraft weight', units='N')
            .add_output('v_stall', shape=(nn,), desc='stall speed', units='m/s')
            .add_output('v_over_v_stall', shape=(nn,), desc='stall speed ratio', units=None)
            )

    if flight_mode == 'runway':
        meta.add_input('mu_r', val=0.05, desc='runway friction coefficient', units=None)
    else:
        meta.add_input('gam', shape=(nn,), desc='flight path angle', units='rad')
        meta.add_output('gam_dot', shape=(nn,), desc='rate of change of flight path angle',
                        units='rad/s', tags=['dymos.state_rate_source:gam'])
        meta.add_output('h_dot', shape=(nn,), desc='rate of change of altitude', units='m/s',
                        tags=['dymos.state_rate_source:h'])

    meta.declare_coloring('*', method=grad_method)
    meta.declare_partials(of='*', wrt='*', method=grad_method)

    return om.ExplicitFuncComp(meta, use_jax=grad_method == 'jax', use_jit=jax_jit)


@use_tempdirs
class TestBalancedFieldFuncComp(unittest.TestCase):

    @unittest.skipIf(jax is None, 'requires jax and jaxlib')
    @require_pyoptsparse('IPOPT')
    def test_balanced_field_func_comp_radau(self):
        self._run_problem(dm.Radau)

    @unittest.skipIf(jax is None, 'requires jax and jaxlib')
    @require_pyoptsparse('IPOPT')
    def test_balanced_field_func_comp_gl(self):
        self._run_problem(dm.GaussLobatto)

    def _run_problem(self, tx):
        p = om.Problem()

        p.driver = om.pyOptSparseDriver()
        p.driver.declare_coloring()

        # Use IPOPT if available, with fallback to SLSQP
        p.driver.options['optimizer'] = 'IPOPT'
        p.driver.options['print_results'] = True

        p.driver.opt_settings['print_level'] = 0
        p.driver.opt_settings['mu_strategy'] = 'adaptive'

        p.driver.opt_settings['bound_mult_init_method'] = 'mu-based'
        p.driver.opt_settings['mu_init'] = 0.01
        p.driver.opt_settings['nlp_scaling_method'] = 'gradient-based'

        # First Phase: Brake release to V1 - both engines operable
        br_to_v1 = dm.Phase(ode_class=wrap_ode_func, transcription=tx(num_segments=3),
                            ode_init_kwargs={'flight_mode': 'runway'})
        br_to_v1.set_time_options(fix_initial=True, duration_bounds=(1, 1000), duration_ref=10.0)
        br_to_v1.add_state('r', fix_initial=True, lower=0, ref=1000.0, defect_ref=1000.0)
        br_to_v1.add_state('v', fix_initial=True, lower=0, ref=100.0, defect_ref=100.0)
        br_to_v1.add_parameter('alpha', val=0.0, opt=False, units='deg')
        br_to_v1.add_timeseries_output('*')

        # Second Phase: Rejected takeoff at V1 - no engines operable
        rto = dm.Phase(ode_class=wrap_ode_func, transcription=tx(num_segments=3),
                       ode_init_kwargs={'flight_mode': 'runway'})
        rto.set_time_options(fix_initial=False, duration_bounds=(1, 1000), duration_ref=1.0)
        rto.add_state('r', fix_initial=False, lower=0, ref=1000.0, defect_ref=1000.0)
        rto.add_state('v', fix_initial=False, lower=0, ref=100.0, defect_ref=100.0)
        rto.add_parameter('alpha', val=0.0, opt=False, units='deg')
        rto.add_timeseries_output('*')

        # Third Phase: V1 to Vr - single engine operable
        v1_to_vr = dm.Phase(ode_class=wrap_ode_func, transcription=tx(num_segments=3),
                            ode_init_kwargs={'flight_mode': 'runway'})
        v1_to_vr.set_time_options(fix_initial=False, duration_bounds=(1, 1000), duration_ref=1.0)
        v1_to_vr.add_state('r', fix_initial=False, lower=0, ref=1000.0, defect_ref=1000.0)
        v1_to_vr.add_state('v', fix_initial=False, lower=0, ref=100.0, defect_ref=100.0)
        v1_to_vr.add_parameter('alpha', val=0.0, opt=False, units='deg')
        v1_to_vr.add_timeseries_output('*')

        # Fourth Phase: Rotate - single engine operable
        rotate = dm.Phase(ode_class=wrap_ode_func, transcription=tx(num_segments=3),
                          ode_init_kwargs={'flight_mode': 'runway'})
        rotate.set_time_options(fix_initial=False, duration_bounds=(1.0, 5), duration_ref=1.0)
        rotate.add_state('r', fix_initial=False, lower=0, ref=1000.0, defect_ref=1000.0)
        rotate.add_state('v', fix_initial=False, lower=0, ref=100.0, defect_ref=100.0)
        rotate.add_polynomial_control('alpha', order=1, opt=True, units='deg', lower=0, upper=10, ref=10, val=[0, 10])
        rotate.add_timeseries_output('*')

        # Fifth Phase: Climb to target speed and altitude at end of runway.
        climb = dm.Phase(ode_class=wrap_ode_func, transcription=tx(num_segments=5),
                         ode_init_kwargs={'flight_mode': 'climb'})
        climb.set_time_options(fix_initial=False, duration_bounds=(1, 100), duration_ref=1.0)
        climb.add_state('r', fix_initial=False, lower=0, ref=1000.0, defect_ref=1000.0)
        climb.add_state('h', fix_initial=True, lower=0, ref=1.0, defect_ref=1.0)
        climb.add_state('v', fix_initial=False, lower=0, ref=100.0, defect_ref=100.0)
        climb.add_state('gam', fix_initial=True, lower=0, ref=0.05, defect_ref=0.05)
        climb.add_control('alpha', opt=True, units='deg', lower=-10, upper=15, ref=10)
        climb.add_timeseries_output('*')

        # Instantiate the trajectory and add phases
        traj = dm.Trajectory()
        p.model.add_subsystem('traj', traj)
        traj.add_phase('br_to_v1', br_to_v1)
        traj.add_phase('rto', rto)
        traj.add_phase('v1_to_vr', v1_to_vr)
        traj.add_phase('rotate', rotate)
        traj.add_phase('climb', climb)

        all_phases = ['br_to_v1', 'v1_to_vr', 'rto', 'rotate', 'climb']
        groundroll_phases = ['br_to_v1', 'v1_to_vr', 'rto', 'rotate']

        # Add parameters common to multiple phases to the trajectory
        traj.add_parameter('m', val=174200., opt=False, units='lbm',
                           desc='aircraft mass',
                           targets={phase: ['m'] for phase in all_phases})

        # Handle parameters which change from phase to phase.
        traj.add_parameter('T_nominal', val=27000 * 2, opt=False, units='lbf', static_target=True,
                           desc='nominal aircraft thrust',
                           targets={'br_to_v1': ['T']})

        traj.add_parameter('T_engine_out', val=27000, opt=False, units='lbf', static_target=True,
                           desc='thrust under a single engine',
                           targets={'v1_to_vr': ['T'], 'rotate': ['T'], 'climb': ['T']})

        traj.add_parameter('T_shutdown', val=0.0, opt=False, units='lbf', static_target=True,
                           desc='thrust when engines are shut down for rejected takeoff',
                           targets={'rto': ['T']})

        traj.add_parameter('mu_r_nominal', val=0.03, opt=False, units=None, static_target=True,
                           desc='nominal runway friction coefficient',
                           targets={'br_to_v1': ['mu_r'], 'v1_to_vr': ['mu_r'],  'rotate': ['mu_r']})

        traj.add_parameter('mu_r_braking', val=0.3, opt=False, units=None, static_target=True,
                           desc='runway friction coefficient under braking',
                           targets={'rto': ['mu_r']})

        traj.add_parameter('h_runway', val=0., opt=False, units='ft',
                           desc='runway altitude',
                           targets={phase: ['h'] for phase in groundroll_phases})

        # Here we're omitting some constants that are common throughout all phases for the sake of brevity.
        # Their correct defaults are specified in add_input calls to `wrap_ode_func`.

        # Standard "end of first phase to beginning of second phase" linkages
        # Alpha changes from being a parameter in v1_to_vr to a polynomial control
        # in rotate, to a dynamic control in `climb`.
        traj.link_phases(['br_to_v1', 'v1_to_vr'], vars=['time', 'r', 'v'])
        traj.link_phases(['v1_to_vr', 'rotate'], vars=['time', 'r', 'v', 'alpha'])
        traj.link_phases(['rotate', 'climb'], vars=['time', 'r', 'v', 'alpha'])
        traj.link_phases(['br_to_v1', 'rto'], vars=['time', 'r', 'v'])

        # Less common "final value of r must match at ends of two phases".
        traj.add_linkage_constraint(phase_a='rto', var_a='r', loc_a='final',
                                    phase_b='climb', var_b='r', loc_b='final',
                                    ref=1000)

        # Define the constraints and objective for the optimal control problem
        v1_to_vr.add_boundary_constraint('v_over_v_stall', loc='final', lower=1.2, ref=100)

        rto.add_boundary_constraint('v', loc='final', equals=0., ref=100, linear=True)

        rotate.add_boundary_constraint('F_r', loc='final', equals=0, ref=100000)

        climb.add_boundary_constraint('h', loc='final', equals=35, ref=35, units='ft', linear=True)
        climb.add_boundary_constraint('gam', loc='final', equals=5, ref=5, units='deg', linear=True)
        climb.add_path_constraint('gam', lower=0, upper=5, ref=5, units='deg')
        climb.add_boundary_constraint('v_over_v_stall', loc='final', lower=1.25, ref=1.25)

        rto.add_objective('r', loc='final', ref=1000.0)

        #
        # Setup the problem and set the initial guess
        #
        p.setup(check=True)

        p.set_val('traj.br_to_v1.t_initial', 0)
        p.set_val('traj.br_to_v1.t_duration', 35)
        p.set_val('traj.br_to_v1.states:r', br_to_v1.interp('r', [0, 2500.0]))
        p.set_val('traj.br_to_v1.states:v', br_to_v1.interp('v', [0, 100.0]))
        p.set_val('traj.br_to_v1.parameters:alpha', 0, units='deg')

        p.set_val('traj.v1_to_vr.t_initial', 35)
        p.set_val('traj.v1_to_vr.t_duration', 35)
        p.set_val('traj.v1_to_vr.states:r', v1_to_vr.interp('r', [2500, 300.0]))
        p.set_val('traj.v1_to_vr.states:v', v1_to_vr.interp('v', [100, 110.0]))
        p.set_val('traj.v1_to_vr.parameters:alpha', 0.0, units='deg')

        p.set_val('traj.rto.t_initial', 35)
        p.set_val('traj.rto.t_duration', 35)
        p.set_val('traj.rto.states:r', rto.interp('r', [2500, 5000.0]))
        p.set_val('traj.rto.states:v', rto.interp('v', [110, 0]))
        p.set_val('traj.rto.parameters:alpha', 0.0, units='deg')

        p.set_val('traj.rotate.t_initial', 70)
        p.set_val('traj.rotate.t_duration', 5)
        p.set_val('traj.rotate.states:r', rotate.interp('r', [1750, 1800.0]))
        p.set_val('traj.rotate.states:v', rotate.interp('v', [80, 85.0]))
        p.set_val('traj.rotate.polynomial_controls:alpha', 0.0, units='deg')

        p.set_val('traj.climb.t_initial', 75)
        p.set_val('traj.climb.t_duration', 15)
        p.set_val('traj.climb.states:r', climb.interp('r', [5000, 5500.0]), units='ft')
        p.set_val('traj.climb.states:v', climb.interp('v', [160, 170.0]), units='kn')
        p.set_val('traj.climb.states:h', climb.interp('h', [0, 35.0]), units='ft')
        p.set_val('traj.climb.states:gam', climb.interp('gam', [0, 5.0]), units='deg')
        p.set_val('traj.climb.controls:alpha', 5.0, units='deg')

        dm.run_problem(p, run_driver=True, simulate=True)

        assert_near_equal(p.get_val('traj.rto.states:r')[-1], 2197.7, tolerance=0.01)



import numpy as np
import openmdao.api as om
from openmdao.utils.units import unit_conversion

from ...grid_data import GridData
from ....options import options as dymos_options


class GaussLobattoInterleaveComp(om.ExplicitComponent):
    r"""
    Class definition for the GaussLobattoInterleaveComp.

    Provides a contiguous output at all nodes for inputs which are only known at
    state discretiation or collocation nodes.

    Parameters
    ----------
    **kwargs : dict
        Dictionary of optional arguments.
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._no_check_partials = not dymos_options['include_check_partials']

    def initialize(self):
        """
        Declare component options.
        """
        self._varnames = {}
        self.options.declare('grid_data', types=GridData, desc='Container object for grid info')

        # Sources is used internally to map the source of a connection to the timeseries to
        # the corresponding input variable.  This is used to ensure that we don't need to connect
        # the same source to this timeseries multiple times.
        self._sources = {'state_disc': {}, 'col': {}}

        # Used to track conversion factors for instances when one output that relies on an input
        # from another variable has potentially different units
        self._units = {}
        self._conversion_factors = {}

    def add_var(self, name, shape, units, disc_src, col_src):
        """
        Add a variable to be interleaved.

        In general these need to be variables whose values are stored separately for state
        discretization or collocation nodes (such as states or ODE outputs).

        Parameters
        ----------
        name : str
            The name of variable as it should appear in the outputs of the
            component ('interleave_comp.all_values:{name}').
        shape : tuple
            The shape of the variable at each instance in time.
        units : str
            The units of the variable.
        disc_src : str
            The source path of the variable's inputs at the discretization nodes.
        col_src : str
            The source path of the variable's inputs at the collocation nodes.

        Returns
        -------
        bool
            True if the variable was added to the interleave comp, False if not due to it already
            being there.
        """
        if name in self._varnames:
            return False

        num_disc_nodes = self.options['grid_data'].subset_num_nodes['state_disc']
        num_col_nodes = self.options['grid_data'].subset_num_nodes['col']
        num_nodes = self.options['grid_data'].subset_num_nodes['all']
        added_source = False

        size = np.prod(shape)

        self._varnames[name] = {}
        self._varnames[name]['state_disc'] = f'disc_values:{name}'
        self._varnames[name]['col'] = f'col_values:{name}'
        self._varnames[name]['all'] = f'all_values:{name}'

        # Check to see if the given disc source has already been used
        # We'll assume that the col source will be the same as well, no need to check both.
        if disc_src in self._sources['state_disc']:
            self._varnames[name]['state_disc'] = self._sources['state_disc'][disc_src]
            self._varnames[name]['col'] = self._sources['col'][col_src]
            input_units = self._units[self._varnames[name]['state_disc']]
        else:
            self.add_input(
                name=self._varnames[name]['state_disc'],
                shape=(num_disc_nodes,) + shape,
                desc=f'Values of {name} at discretization nodes',
                units=units)
            self.add_input(
                name=self._varnames[name]['col'],
                shape=(num_col_nodes,) + shape,
                desc=f'Values of {name} at collocation nodes',
                units=units)
            self._sources['state_disc'][disc_src] = self._varnames[name]['state_disc']
            self._sources['col'][col_src] = self._varnames[name]['col']
            input_units = self._units[self._varnames[name]['state_disc']] = units
            added_source = True

        self.add_output(
            name=self._varnames[name]['all'],
            shape=(num_nodes,) + shape,
            desc=f'Values of {name} at all nodes',
            units=units)

        start_rows = self.options['grid_data'].subset_node_indices['state_disc'] * size
        r = (start_rows[:, np.newaxis] + np.arange(size, dtype=int)).ravel()
        c = np.arange(size * num_disc_nodes, dtype=int)

        # There's a chance that the input for this output was pulled from another variable with
        # different units, so account for that with a conversion.
        if input_units is None or units is None:
            scale = 1.0
            offset = 0
        else:
            scale, offset = unit_conversion(input_units, units)
        self._conversion_factors[self._varnames[name]['all']] = scale, offset

        self.declare_partials(of=self._varnames[name]['all'],
                              wrt=self._varnames[name]['state_disc'],
                              rows=r, cols=c, val=scale)

        start_rows = self.options['grid_data'].subset_node_indices['col'] * size
        r = (start_rows[:, np.newaxis] + np.arange(size, dtype=int)).ravel()
        c = np.arange(size * num_col_nodes, dtype=int)

        self.declare_partials(of=self._varnames[name]['all'],
                              wrt=self._varnames[name]['col'],
                              rows=r, cols=c, val=scale)

        return added_source

    def compute(self, inputs, outputs):
        """
        Compute outputs for all nodes.

        Parameters
        ----------
        inputs : `Vector`
            `Vector` containing inputs.
        outputs : `Vector`
            `Vector` containing outputs.
        """
        disc_idxs = self.options['grid_data'].subset_node_indices['state_disc']
        col_idxs = self.options['grid_data'].subset_node_indices['col']

        for name, varnames in self._varnames.items():
            scale, offset = self._conversion_factors[self._varnames[name]['all']]
            outputs[varnames['all']][disc_idxs] = inputs[varnames['state_disc']]
            outputs[varnames['all']][col_idxs] = inputs[varnames['col']]
            outputs[varnames['all']] *= scale
            outputs[varnames['all']] += offset



import unittest


from openmdao.utils.assert_utils import assert_near_equal
from openmdao.utils.testing_utils import use_tempdirs, require_pyoptsparse


@use_tempdirs
class TestDocSSTOEarth(unittest.TestCase):

    @require_pyoptsparse(optimizer='SLSQP')
    def test_doc_ssto_earth(self):
        import matplotlib.pyplot as plt
        import openmdao.api as om
        import dymos as dm

        #
        # Setup and solve the optimal control problem
        #
        p = om.Problem(model=om.Group())
        p.driver = om.pyOptSparseDriver()
        p.driver.declare_coloring(tol=1.0E-12)

        from dymos.examples.ssto.launch_vehicle_ode import LaunchVehicleODE

        #
        # Initialize our Trajectory and Phase
        #
        traj = dm.Trajectory()

        phase = dm.Phase(ode_class=LaunchVehicleODE,
                         transcription=dm.GaussLobatto(num_segments=12, order=3, compressed=False))

        traj.add_phase('phase0', phase)
        p.model.add_subsystem('traj', traj)

        #
        # Set the options for the variables
        #
        phase.set_time_options(fix_initial=True, duration_bounds=(10, 500))

        phase.add_state('x', fix_initial=True, ref=1.0E5, defect_ref=10000.0,
                        rate_source='xdot')
        phase.add_state('y', fix_initial=True, ref=1.0E5, defect_ref=10000.0,
                        rate_source='ydot')
        phase.add_state('vx', fix_initial=True, ref=1.0E3, defect_ref=1000.0,
                        rate_source='vxdot')
        phase.add_state('vy', fix_initial=True, ref=1.0E3, defect_ref=1000.0,
                        rate_source='vydot')
        phase.add_state('m', fix_initial=True, ref=1.0E3, defect_ref=100.0,
                        rate_source='mdot')

        phase.add_control('theta', units='rad', lower=-1.57, upper=1.57, targets=['theta'])
        phase.add_parameter('thrust', units='N', opt=False, val=2100000.0, targets=['thrust'])

        #
        # Set the options for our constraints and objective
        #
        phase.add_boundary_constraint('y', loc='final', equals=1.85E5, linear=True)
        phase.add_boundary_constraint('vx', loc='final', equals=7796.6961)
        phase.add_boundary_constraint('vy', loc='final', equals=0)

        phase.add_objective('time', loc='final', scaler=0.01)

        p.model.linear_solver = om.DirectSolver()

        #
        # Setup and set initial values
        #
        p.setup(check=True)

        p.set_val('traj.phase0.t_initial', 0.0)
        p.set_val('traj.phase0.t_duration', 150.0)
        p.set_val('traj.phase0.states:x', phase.interp('x', [0, 1.15E5]))
        p.set_val('traj.phase0.states:y', phase.interp('y', [0, 1.85E5]))
        p.set_val('traj.phase0.states:vy', phase.interp('vx', [1.0E-6, 0]))
        p.set_val('traj.phase0.states:m', phase.interp('vy', [117000, 1163]))
        p.set_val('traj.phase0.controls:theta', phase.interp('theta', [1.5, -0.76]))
        p.set_val('traj.phase0.parameters:thrust', 2.1, units='MN')

        #
        # Solve the Problem
        #
        dm.run_problem(p)

        assert_near_equal(p.get_val('traj.phase0.timeseries.time')[-1], 143, tolerance=0.05)
        assert_near_equal(p.get_val('traj.phase0.timeseries.states:y')[-1], 1.85E5, 1e-4)
        assert_near_equal(p.get_val('traj.phase0.timeseries.states:vx')[-1], 7796.6961, 1e-4)
        assert_near_equal(p.get_val('traj.phase0.timeseries.states:vy')[-1], 0, 1e-4)
        #
        # Get the explicitly simulated results
        #
        exp_out = traj.simulate()

        #
        # Plot the results
        #
        fig, axes = plt.subplots(nrows=2, ncols=1, figsize=(10, 8))

        axes[0].plot(p.get_val('traj.phase0.timeseries.states:x'),
                     p.get_val('traj.phase0.timeseries.states:y'),
                     marker='o',
                     ms=4,
                     linestyle='None',
                     label='solution')

        axes[0].plot(exp_out.get_val('traj.phase0.timeseries.states:x'),
                     exp_out.get_val('traj.phase0.timeseries.states:y'),
                     marker=None,
                     linestyle='-',
                     label='simulation')

        axes[0].set_xlabel('range (m)')
        axes[0].set_ylabel('altitude (m)')
        axes[0].set_aspect('equal')

        axes[1].plot(p.get_val('traj.phase0.timeseries.time'),
                     p.get_val('traj.phase0.timeseries.controls:theta'),
                     marker='o',
                     ms=4,
                     linestyle='None')

        axes[1].plot(exp_out.get_val('traj.phase0.timeseries.time'),
                     exp_out.get_val('traj.phase0.timeseries.controls:theta'),
                     linestyle='-',
                     marker=None)

        axes[1].set_xlabel('time (s)')
        axes[1].set_ylabel('theta (deg)')

        plt.suptitle('Single Stage to Orbit Solution Using Linear Tangent Guidance')
        fig.legend(loc='lower center', ncol=2)

        plt.show()


if __name__ == "__main__":
    unittest.main()



import unittest

import numpy as np

import matplotlib.pyplot as plt
plt.switch_backend('Agg')

import openmdao.api as om
from openmdao.utils.assert_utils import assert_near_equal
from openmdao.utils.testing_utils import use_tempdirs

import dymos as dm
from dymos.examples.brachistochrone.brachistochrone_vector_states_ode \
    import BrachistochroneVectorStatesODE

SHOW_PLOTS = True


@use_tempdirs
class TestBrachistochroneVectorPathConstraints(unittest.TestCase):

    def test_brachistochrone_vector_state_path_constraints_radau_partial_indices(self):

        p = om.Problem(model=om.Group())

        p.driver = om.ScipyOptimizeDriver()
        p.driver.declare_coloring()

        phase = dm.Phase(ode_class=BrachistochroneVectorStatesODE,
                         transcription=dm.Radau(num_segments=20, order=3))

        p.model.add_subsystem('phase0', phase)

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

        phase.add_state('pos', fix_initial=True, fix_final=True)
        phase.add_state('v', fix_initial=True, fix_final=False)

        phase.add_control('theta', units='deg', rate_continuity=True, lower=0.01, upper=179.9)

        phase.add_parameter('g', units='m/s**2', opt=False, val=9.80665)

        phase.add_boundary_constraint('theta_rate', loc='final', equals=0.0)
        phase.add_boundary_constraint('theta_rate2', loc='final', equals=0.0)
        phase.add_path_constraint('pos', indices=[1], lower=5)

        phase.add_timeseries_output('pos_dot')

        # Minimize time at the end of the phase
        phase.add_objective('time', loc='final', scaler=10)

        p.model.linear_solver = om.DirectSolver()
        p.setup(check=True, force_alloc_complex=True)

        p['phase0.t_initial'] = 0.0
        p['phase0.t_duration'] = 2.0

        pos0 = [0, 10]
        posf = [10, 5]

        p['phase0.states:pos'] = phase.interp('pos', [pos0, posf])
        p['phase0.states:v'] = phase.interp('v', [0, 9.9])
        p['phase0.controls:theta'] = phase.interp('theta', [5, 100])
        p['phase0.parameters:g'] = 9.80665

        p.run_driver()

        assert_near_equal(np.min(p.get_val('phase0.timeseries.states:pos')[:, 1]), 5.0,
                          tolerance=1.0E-3)

        # Plot results
        if SHOW_PLOTS:
            exp_out = phase.simulate(times_per_seg=10)

            fig, ax = plt.subplots()
            fig.suptitle('Brachistochrone Solution')

            x_imp = p.get_val('phase0.timeseries.states:pos')[:, 0]
            y_imp = p.get_val('phase0.timeseries.states:pos')[:, 1]

            x_exp = exp_out.get_val('phase0.timeseries.states:pos')[:, 0]
            y_exp = exp_out.get_val('phase0.timeseries.states:pos')[:, 1]

            ax.plot(x_imp, y_imp, 'ro', label='implicit')
            ax.plot(x_exp, y_exp, 'b-', label='explicit')

            ax.set_xlabel('x (m)')
            ax.set_ylabel('y (m)')
            ax.grid(True)
            ax.legend(loc='upper right')

            fig, ax = plt.subplots()
            fig.suptitle('Brachistochrone Solution\nVelocity')

            t_imp = p.get_val('phase0.timeseries.time')
            t_exp = exp_out.get_val('phase0.timeseries.time')

            xdot_imp = p.get_val('phase0.timeseries.pos_dot')[:, 0]
            ydot_imp = p.get_val('phase0.timeseries.pos_dot')[:, 1]

            xdot_exp = exp_out.get_val('phase0.timeseries.pos_dot')[:, 0]
            ydot_exp = exp_out.get_val('phase0.timeseries.pos_dot')[:, 1]

            ax.plot(t_imp, xdot_imp, 'bo', label='implicit')
            ax.plot(t_exp, xdot_exp, 'b-', label='explicit')

            ax.plot(t_imp, ydot_imp, 'ro', label='implicit')
            ax.plot(t_exp, ydot_exp, 'r-', label='explicit')

            ax.set_xlabel('t (s)')
            ax.set_ylabel('v (m/s)')
            ax.grid(True)
            ax.legend(loc='upper right')

            fig, ax = plt.subplots()
            fig.suptitle('Brachistochrone Solution')

            x_imp = p.get_val('phase0.timeseries.time')
            y_imp = p.get_val('phase0.timeseries.control_rates:theta_rate2')

            x_exp = exp_out.get_val('phase0.timeseries.time')
            y_exp = exp_out.get_val('phase0.timeseries.control_rates:theta_rate2')

            ax.plot(x_imp, y_imp, 'ro', label='implicit')
            ax.plot(x_exp, y_exp, 'b-', label='explicit')

            ax.set_xlabel('time (s)')
            ax.set_ylabel('theta rate2 (rad/s**2)')
            ax.grid(True)
            ax.legend(loc='lower right')

            plt.show()

        return p

    def test_brachistochrone_vector_ode_path_constraints_radau_partial_indices(self):

        p = om.Problem(model=om.Group())

        p.driver = om.ScipyOptimizeDriver()
        p.driver.declare_coloring()

        phase = dm.Phase(ode_class=BrachistochroneVectorStatesODE,
                         transcription=dm.Radau(num_segments=20, order=3))

        p.model.add_subsystem('phase0', phase)

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

        phase.add_state('pos', fix_initial=True, fix_final=True)
        phase.add_state('v', fix_initial=True, fix_final=False)

        phase.add_control('theta', units='deg', rate_continuity=True, lower=0.01, upper=179.9)

        phase.add_parameter('g', units='m/s**2', opt=False, val=9.80665)

        phase.add_path_constraint('pos_dot', indices=[1],
                                  lower=-4, upper=4)

        # Minimize time at the end of the phase
        phase.add_objective('time', loc='final', scaler=10)

        p.model.linear_solver = om.DirectSolver()
        p.setup(check=True, force_alloc_complex=True)

        p['phase0.t_initial'] = 0.0
        p['phase0.t_duration'] = 2.0

        pos0 = [0, 10]
        posf = [10, 5]

        p['phase0.states:pos'] = phase.interp('pos', [pos0, posf])
        p['phase0.states:v'] = phase.interp('v', [0, 9.9])
        p['phase0.controls:theta'] = phase.interp('theta', [5, 100])
        p['phase0.parameters:g'] = 9.80665

        p.run_driver()

        assert_near_equal(np.min(p.get_val('phase0.timeseries.pos_dot')[:, -1]),
                          -4,
                          tolerance=1.0E-2)

        # Plot results
        if SHOW_PLOTS:
            exp_out = phase.simulate(times_per_seg=20)

            fig, ax = plt.subplots()
            fig.suptitle('Brachistochrone Solution')

            x_imp = p.get_val('phase0.timeseries.states:pos')[:, 0]
            y_imp = p.get_val('phase0.timeseries.states:pos')[:, 1]

            x_exp = exp_out.get_val('phase0.timeseries.states:pos')[:, 0]
            y_exp = exp_out.get_val('phase0.timeseries.states:pos')[:, 1]

            ax.plot(x_imp, y_imp, 'ro', label='implicit')
            ax.plot(x_exp, y_exp, 'b-', label='explicit')

            ax.set_xlabel('x (m)')
            ax.set_ylabel('y (m)')
            ax.grid(True)
            ax.legend(loc='upper right')

            fig, ax = plt.subplots()
            fig.suptitle('Brachistochrone Solution\nVelocity')

            t_imp = p.get_val('phase0.timeseries.time')
            t_exp = exp_out.get_val('phase0.timeseries.time')

            xdot_imp = p.get_val('phase0.timeseries.pos_dot')[:, 0]
            ydot_imp = p.get_val('phase0.timeseries.pos_dot')[:, 1]

            xdot_exp = exp_out.get_val('phase0.timeseries.pos_dot')[:, 0]
            ydot_exp = exp_out.get_val('phase0.timeseries.pos_dot')[:, 1]

            ax.plot(t_imp, xdot_imp, 'bo', label='implicit')
            ax.plot(t_exp, xdot_exp, 'b-', label='explicit')

            ax.plot(t_imp, ydot_imp, 'ro', label='implicit')
            ax.plot(t_exp, ydot_exp, 'r-', label='explicit')

            ax.set_xlabel('t (s)')
            ax.set_ylabel('v (m/s)')
            ax.grid(True)
            ax.legend(loc='upper right')

            fig, ax = plt.subplots()
            fig.suptitle('Brachistochrone Solution')

            x_imp = p.get_val('phase0.timeseries.time')
            y_imp = p.get_val('phase0.timeseries.control_rates:theta_rate2')

            x_exp = exp_out.get_val('phase0.timeseries.time')
            y_exp = exp_out.get_val('phase0.timeseries.control_rates:theta_rate2')

            ax.plot(x_imp, y_imp, 'ro', label='implicit')
            ax.plot(x_exp, y_exp, 'b-', label='explicit')

            ax.set_xlabel('time (s)')
            ax.set_ylabel('theta rate2 (rad/s**2)')
            ax.grid(True)
            ax.legend(loc='lower right')

            plt.show()

        return p

    def test_brachistochrone_vector_ode_path_constraints_radau_no_indices(self):

        p = om.Problem(model=om.Group())

        p.driver = om.ScipyOptimizeDriver()
        p.driver.declare_coloring()

        phase = dm.Phase(ode_class=BrachistochroneVectorStatesODE,
                         transcription=dm.Radau(num_segments=20, order=3))

        p.model.add_subsystem('phase0', phase)

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

        phase.add_state('pos', fix_initial=True, fix_final=True)
        phase.add_state('v', fix_initial=True, fix_final=False)

        phase.add_control('theta', units='deg', rate_continuity=True, lower=0.01, upper=179.9)

        phase.add_parameter('g', units='m/s**2', opt=False, val=9.80665)

        phase.add_path_constraint('pos_dot', lower=-4, upper=12)

        phase.add_timeseries_output('pos_dot')

        # Minimize time at the end of the phase
        phase.add_objective('time', loc='final', scaler=10)

        p.model.linear_solver = om.DirectSolver()
        p.setup(check=True, force_alloc_complex=True)

        p['phase0.t_initial'] = 0.0
        p['phase0.t_duration'] = 2.0

        pos0 = [0, 10]
        posf = [10, 5]

        p['phase0.states:pos'] = phase.interp('pos', [pos0, posf])
        p['phase0.states:v'] = phase.interp('v', [0, 9.9])
        p['phase0.controls:theta'] = phase.interp('theta', [5, 100])
        p['phase0.parameters:g'] = 9.80665

        p.run_driver()

        assert_near_equal(np.min(p.get_val('phase0.timeseries.pos_dot')[:, -1]),
                          -4,
                          tolerance=1.0E-2)

        # Plot results
        if SHOW_PLOTS:
            exp_out = phase.simulate(times_per_seg=50)

            fig, ax = plt.subplots()
            fig.suptitle('Brachistochrone Solution')

            x_imp = p.get_val('phase0.timeseries.states:pos')[:, 0]
            y_imp = p.get_val('phase0.timeseries.states:pos')[:, 1]

            x_exp = exp_out.get_val('phase0.timeseries.states:pos')[:, 0]
            y_exp = exp_out.get_val('phase0.timeseries.states:pos')[:, 1]

            ax.plot(x_imp, y_imp, 'ro', label='implicit')
            ax.plot(x_exp, y_exp, 'b-', label='explicit')

            ax.set_xlabel('x (m)')
            ax.set_ylabel('y (m)')
            ax.grid(True)
            ax.legend(loc='upper right')

            fig, ax = plt.subplots()
            fig.suptitle('Brachistochrone Solution\nVelocity')

            t_imp = p.get_val('phase0.timeseries.time')
            t_exp = exp_out.get_val('phase0.timeseries.time')

            xdot_imp = p.get_val('phase0.timeseries.pos_dot')[:, 0]
            ydot_imp = p.get_val('phase0.timeseries.pos_dot')[:, 1]

            xdot_exp = exp_out.get_val('phase0.timeseries.pos_dot')[:, 0]
            ydot_exp = exp_out.get_val('phase0.timeseries.pos_dot')[:, 1]

            ax.plot(t_imp, xdot_imp, 'bo', label='implicit')
            ax.plot(t_exp, xdot_exp, 'b-', label='explicit')

            ax.plot(t_imp, ydot_imp, 'ro', label='implicit')
            ax.plot(t_exp, ydot_exp, 'r-', label='explicit')

            ax.set_xlabel('t (s)')
            ax.set_ylabel('v (m/s)')
            ax.grid(True)
            ax.legend(loc='upper right')

            fig, ax = plt.subplots()
            fig.suptitle('Brachistochrone Solution')

            x_imp = p.get_val('phase0.timeseries.time')
            y_imp = p.get_val('phase0.timeseries.control_rates:theta_rate2')

            x_exp = exp_out.get_val('phase0.timeseries.time')
            y_exp = exp_out.get_val('phase0.timeseries.control_rates:theta_rate2')

            ax.plot(x_imp, y_imp, 'ro', label='implicit')
            ax.plot(x_exp, y_exp, 'b-', label='explicit')

            ax.set_xlabel('time (s)')
            ax.set_ylabel('theta rate2 (rad/s**2)')
            ax.grid(True)
            ax.legend(loc='lower right')

            plt.show()

        return p

    def test_brachistochrone_vector_state_path_constraints_gl_partial_indices(self):

        p = om.Problem(model=om.Group())

        p.driver = om.ScipyOptimizeDriver()
        p.driver.declare_coloring()

        phase = dm.Phase(ode_class=BrachistochroneVectorStatesODE,
                         transcription=dm.GaussLobatto(num_segments=20, order=3))

        p.model.add_subsystem('phase0', phase)

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

        phase.add_state('pos', fix_initial=True, fix_final=True)
        phase.add_state('v', fix_initial=True, fix_final=False)

        phase.add_control('theta', units='deg', rate_continuity=True, lower=0.01, upper=179.9)

        phase.add_parameter('g', units='m/s**2', opt=False, val=9.80665)

        phase.add_path_constraint('pos', indices=[1], lower=5)

        phase.add_timeseries_output('pos_dot')

        # Minimize time at the end of the phase
        phase.add_objective('time', loc='final', scaler=10)

        p.model.linear_solver = om.DirectSolver()
        p.setup(check=True, force_alloc_complex=True)

        p['phase0.t_initial'] = 0.0
        p['phase0.t_duration'] = 2.0

        pos0 = [0, 10]
        posf = [10, 5]

        p['phase0.states:pos'] = phase.interp('pos', [pos0, posf])
        p['phase0.states:v'] = phase.interp('v', [0, 9.9])
        p['phase0.controls:theta'] = phase.interp('theta', [5, 100])
        p['phase0.parameters:g'] = 9.80665

        p.run_driver()

        assert_near_equal(np.min(p.get_val('phase0.timeseries.states:pos')[:, 1]),
                          5,
                          tolerance=1.0E-2)

        # Plot results
        if SHOW_PLOTS:
            exp_out = phase.simulate(times_per_seg=20)

            fig, ax = plt.subplots()
            fig.suptitle('Brachistochrone Solution')

            x_imp = p.get_val('phase0.timeseries.states:pos')[:, 0]
            y_imp = p.get_val('phase0.timeseries.states:pos')[:, 1]

            x_exp = exp_out.get_val('phase0.timeseries.states:pos')[:, 0]
            y_exp = exp_out.get_val('phase0.timeseries.states:pos')[:, 1]

            ax.plot(x_imp, y_imp, 'ro', label='implicit')
            ax.plot(x_exp, y_exp, 'b-', label='explicit')

            ax.set_xlabel('x (m)')
            ax.set_ylabel('y (m)')
            ax.grid(True)
            ax.legend(loc='upper right')

            fig, ax = plt.subplots()
            fig.suptitle('Brachistochrone Solution\nVelocity')

            t_imp = p.get_val('phase0.timeseries.time')
            t_exp = exp_out.get_val('phase0.timeseries.time')

            xdot_imp = p.get_val('phase0.timeseries.pos_dot')[:, 0]
            ydot_imp = p.get_val('phase0.timeseries.pos_dot')[:, 1]

            xdot_exp = exp_out.get_val('phase0.timeseries.pos_dot')[:, 0]
            ydot_exp = exp_out.get_val('phase0.timeseries.pos_dot')[:, 1]

            ax.plot(t_imp, xdot_imp, 'bo', label='implicit')
            ax.plot(t_exp, xdot_exp, 'b-', label='explicit')

            ax.plot(t_imp, ydot_imp, 'ro', label='implicit')
            ax.plot(t_exp, ydot_exp, 'r-', label='explicit')

            ax.set_xlabel('t (s)')
            ax.set_ylabel('v (m/s)')
            ax.grid(True)
            ax.legend(loc='upper right')

            fig, ax = plt.subplots()
            fig.suptitle('Brachistochrone Solution')

            x_imp = p.get_val('phase0.timeseries.time')
            y_imp = p.get_val('phase0.timeseries.control_rates:theta_rate2')

            x_exp = exp_out.get_val('phase0.timeseries.time')
            y_exp = exp_out.get_val('phase0.timeseries.control_rates:theta_rate2')

            ax.plot(x_imp, y_imp, 'ro', label='implicit')
            ax.plot(x_exp, y_exp, 'b-', label='explicit')

            ax.set_xlabel('time (s)')
            ax.set_ylabel('theta rate2 (rad/s**2)')
            ax.grid(True)
            ax.legend(loc='lower right')

            plt.show()

        return p

    def test_brachistochrone_vector_ode_path_constraints_gl_partial_indices(self):

        p = om.Problem(model=om.Group())

        p.driver = om.ScipyOptimizeDriver()
        p.driver.declare_coloring()

        phase = dm.Phase(ode_class=BrachistochroneVectorStatesODE,
                         transcription=dm.GaussLobatto(num_segments=20, order=3))

        p.model.add_subsystem('phase0', phase)

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

        phase.add_state('pos', fix_initial=True, fix_final=True)
        phase.add_state('v', fix_initial=True, fix_final=False)

        phase.add_control('theta', units='deg',
                          rate_continuity=True, lower=0.01, upper=179.9)

        phase.add_parameter('g', units='m/s**2', opt=False, val=9.80665)

        phase.add_boundary_constraint('theta_rate', loc='final', equals=0.0)
        phase.add_boundary_constraint('theta_rate2', loc='final', equals=0.0)
        phase.add_path_constraint('pos_dot', indices=[1],
                                  lower=-4, upper=4)

        phase.add_timeseries_output('pos_dot')

        # Minimize time at the end of the phase
        phase.add_objective('time', loc='final', scaler=10)

        p.model.linear_solver = om.DirectSolver()
        p.setup(check=True, force_alloc_complex=True)

        p['phase0.t_initial'] = 0.0
        p['phase0.t_duration'] = 2.0

        pos0 = [0, 10]
        posf = [10, 5]

        p['phase0.states:pos'] = phase.interp('pos', [pos0, posf])
        p['phase0.states:v'] = phase.interp('v', [0, 9.9])
        p['phase0.controls:theta'] = phase.interp('theta', [5, 100])
        p['phase0.parameters:g'] = 9.80665

        p.run_driver()

        assert_near_equal(np.min(p.get_val('phase0.timeseries.pos_dot')[:, 1]), -4.0,
                          tolerance=1.0E-3)

        # Plot results
        if SHOW_PLOTS:
            exp_out = phase.simulate(times_per_seg=20)

            fig, ax = plt.subplots()
            fig.suptitle('Brachistochrone Solution')

            x_imp = p.get_val('phase0.timeseries.states:pos')[:, 0]
            y_imp = p.get_val('phase0.timeseries.states:pos')[:, 1]

            x_exp = exp_out.get_val('phase0.timeseries.states:pos')[:, 0]
            y_exp = exp_out.get_val('phase0.timeseries.states:pos')[:, 1]

            ax.plot(x_imp, y_imp, 'ro', label='implicit')
            ax.plot(x_exp, y_exp, 'b-', label='explicit')

            ax.set_xlabel('x (m)')
            ax.set_ylabel('y (m)')
            ax.grid(True)
            ax.legend(loc='upper right')

            fig, ax = plt.subplots()
            fig.suptitle('Brachistochrone Solution\nVelocity')

            t_imp = p.get_val('phase0.timeseries.time')
            t_exp = exp_out.get_val('phase0.timeseries.time')

            xdot_imp = p.get_val('phase0.timeseries.pos_dot')[:, 0]
            ydot_imp = p.get_val('phase0.timeseries.pos_dot')[:, 1]

            xdot_exp = exp_out.get_val('phase0.timeseries.pos_dot')[:, 0]
            ydot_exp = exp_out.get_val('phase0.timeseries.pos_dot')[:, 1]

            ax.plot(t_imp, xdot_imp, 'bo', label='implicit')
            ax.plot(t_exp, xdot_exp, 'b-', label='explicit')

            ax.plot(t_imp, ydot_imp, 'ro', label='implicit')
            ax.plot(t_exp, ydot_exp, 'r-', label='explicit')

            ax.set_xlabel('t (s)')
            ax.set_ylabel('v (m/s)')
            ax.grid(True)
            ax.legend(loc='upper right')

            fig, ax = plt.subplots()
            fig.suptitle('Brachistochrone Solution')

            x_imp = p.get_val('phase0.timeseries.time')
            y_imp = p.get_val('phase0.timeseries.control_rates:theta_rate2')

            x_exp = exp_out.get_val('phase0.timeseries.time')
            y_exp = exp_out.get_val('phase0.timeseries.control_rates:theta_rate2')

            ax.plot(x_imp, y_imp, 'ro', label='implicit')
            ax.plot(x_exp, y_exp, 'b-', label='explicit')

            ax.set_xlabel('time (s)')
            ax.set_ylabel('theta rate2 (rad/s**2)')
            ax.grid(True)
            ax.legend(loc='lower right')

            plt.show()

        return p

    def test_brachistochrone_vector_ode_path_constraints_gl_no_indices(self):

        p = om.Problem(model=om.Group())

        p.driver = om.ScipyOptimizeDriver()
        p.driver.declare_coloring()

        phase = dm.Phase(ode_class=BrachistochroneVectorStatesODE,
                         transcription=dm.GaussLobatto(num_segments=20, order=3))

        p.model.add_subsystem('phase0', phase)

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

        phase.add_state('pos', fix_initial=True, fix_final=True)
        phase.add_state('v', fix_initial=True, fix_final=False)

        phase.add_control('theta', units='deg',
                          rate_continuity=False, lower=0.01, upper=179.9)

        phase.add_parameter('g', units='m/s**2', opt=False, val=9.80665)

        phase.add_path_constraint('pos_dot', lower=-4, upper=12)

        phase.add_timeseries_output('pos_dot')

        # Minimize time at the end of the phase
        phase.add_objective('time', loc='final', scaler=10)

        p.model.linear_solver = om.DirectSolver()
        p.setup(check=True, force_alloc_complex=True)

        p['phase0.t_initial'] = 0.0
        p['phase0.t_duration'] = 2.0

        pos0 = [0, 10]
        posf = [10, 5]

        p['phase0.states:pos'] = phase.interp('pos', [pos0, posf])
        p['phase0.states:v'] = phase.interp('v', [0, 9.9])
        p['phase0.controls:theta'] = phase.interp('theta', [5, 100])
        p['phase0.parameters:g'] = 9.80665

        p.run_driver()

        assert_near_equal(np.min(p.get_val('phase0.timeseries.pos_dot')[:, -1]),
                          -4,
                          tolerance=1.0E-2)

        # Plot results
        if SHOW_PLOTS:
            exp_out = phase.simulate(times_per_seg=20)

            fig, ax = plt.subplots()
            fig.suptitle('Brachistochrone Solution')

            x_imp = p.get_val('phase0.timeseries.states:pos')[:, 0]
            y_imp = p.get_val('phase0.timeseries.states:pos')[:, 1]

            x_exp = exp_out.get_val('phase0.timeseries.states:pos')[:, 0]
            y_exp = exp_out.get_val('phase0.timeseries.states:pos')[:, 1]

            ax.plot(x_imp, y_imp, 'ro', label='implicit')
            ax.plot(x_exp, y_exp, 'b-', label='explicit')

            ax.set_xlabel('x (m)')
            ax.set_ylabel('y (m)')
            ax.grid(True)
            ax.legend(loc='upper right')

            fig, ax = plt.subplots()
            fig.suptitle('Brachistochrone Solution\nVelocity')

            t_imp = p.get_val('phase0.timeseries.time')
            t_exp = exp_out.get_val('phase0.timeseries.time')

            xdot_imp = p.get_val('phase0.timeseries.pos_dot')[:, 0]
            ydot_imp = p.get_val('phase0.timeseries.pos_dot')[:, 1]

            xdot_exp = exp_out.get_val('phase0.timeseries.pos_dot')[:, 0]
            ydot_exp = exp_out.get_val('phase0.timeseries.pos_dot')[:, 1]

            ax.plot(t_imp, xdot_imp, 'bo', label='implicit')
            ax.plot(t_exp, xdot_exp, 'b-', label='explicit')

            ax.plot(t_imp, ydot_imp, 'ro', label='implicit')
            ax.plot(t_exp, ydot_exp, 'r-', label='explicit')

            ax.set_xlabel('t (s)')
            ax.set_ylabel('v (m/s)')
            ax.grid(True)
            ax.legend(loc='upper right')

            fig, ax = plt.subplots()
            fig.suptitle('Brachistochrone Solution')

            x_imp = p.get_val('phase0.timeseries.time')
            y_imp = p.get_val('phase0.timeseries.control_rates:theta_rate2')

            x_exp = exp_out.get_val('phase0.timeseries.time')
            y_exp = exp_out.get_val('phase0.timeseries.control_rates:theta_rate2')

            ax.plot(x_imp, y_imp, 'ro', label='implicit')
            ax.plot(x_exp, y_exp, 'b-', label='explicit')

            ax.set_xlabel('time (s)')
            ax.set_ylabel('theta rate2 (rad/s**2)')
            ax.grid(True)
            ax.legend(loc='lower right')

            plt.show()

        return p


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



import unittest

import matplotlib.pyplot as plt
plt.switch_backend('Agg')
plt.style.use('ggplot')

from openmdao.utils.testing_utils import use_tempdirs, require_pyoptsparse


@use_tempdirs
class TestFiniteBurnOrbitRaise(unittest.TestCase):

    @require_pyoptsparse(optimizer='IPOPT')
    def test_finite_burn_orbit_raise(self):
        import numpy as np
        import matplotlib.pyplot as plt
        import openmdao.api as om
        from openmdao.utils.assert_utils import assert_near_equal

        import dymos as dm
        from dymos.examples.finite_burn_orbit_raise.finite_burn_eom import FiniteBurnODE

        p = om.Problem(model=om.Group())

        p.driver = om.pyOptSparseDriver()
        p.driver.options['optimizer'] = 'IPOPT'
        p.driver.declare_coloring()

        traj = dm.Trajectory()

        traj.add_parameter('c', opt=False, val=1.5, units='DU/TU',
                           targets={'burn1': ['c'], 'coast': ['c'], 'burn2': ['c']})

        # First Phase (burn)

        burn1 = dm.Phase(ode_class=FiniteBurnODE,
                         transcription=dm.GaussLobatto(num_segments=5, order=3, compressed=False))

        burn1 = traj.add_phase('burn1', burn1)

        burn1.set_time_options(fix_initial=True, duration_bounds=(.5, 10), units='TU')
        burn1.add_state('r', fix_initial=True, fix_final=False, defect_scaler=100.0,
                        rate_source='r_dot', units='DU')
        burn1.add_state('theta', fix_initial=True, fix_final=False, defect_scaler=100.0,
                        rate_source='theta_dot', units='rad')
        burn1.add_state('vr', fix_initial=True, fix_final=False, defect_scaler=100.0,
                        rate_source='vr_dot', units='DU/TU')
        burn1.add_state('vt', fix_initial=True, fix_final=False, defect_scaler=100.0,
                        rate_source='vt_dot', units='DU/TU')
        burn1.add_state('accel', fix_initial=True, fix_final=False,
                        rate_source='at_dot', units='DU/TU**2')
        burn1.add_state('deltav', fix_initial=True, fix_final=False,
                        rate_source='deltav_dot', units='DU/TU')
        burn1.add_control('u1', rate_continuity=True, rate2_continuity=True, units='deg',
                          scaler=0.01, rate_continuity_scaler=0.001, rate2_continuity_scaler=0.001,
                          lower=-30, upper=30)
        # Second Phase (Coast)
        coast = dm.Phase(ode_class=FiniteBurnODE,
                         transcription=dm.GaussLobatto(num_segments=5, order=3, compressed=False))

        coast.set_time_options(initial_bounds=(0.5, 20), duration_bounds=(.5, 50), duration_ref=50,
                               units='TU')
        coast.add_state('r', fix_initial=False, fix_final=False, defect_scaler=100.0,
                        rate_source='r_dot', targets=['r'], units='DU')
        coast.add_state('theta', fix_initial=False, fix_final=False, defect_scaler=100.0,
                        rate_source='theta_dot', targets=['theta'], units='rad')
        coast.add_state('vr', fix_initial=False, fix_final=False, defect_scaler=100.0,
                        rate_source='vr_dot', targets=['vr'], units='DU/TU')
        coast.add_state('vt', fix_initial=False, fix_final=False, defect_scaler=100.0,
                        rate_source='vt_dot', targets=['vt'], units='DU/TU')
        coast.add_state('accel', fix_initial=True, fix_final=True,
                        rate_source='at_dot', targets=['accel'], units='DU/TU**2')
        coast.add_state('deltav', fix_initial=False, fix_final=False,
                        rate_source='deltav_dot', units='DU/TU')

        coast.add_parameter('u1', opt=False, val=0.0, units='deg', targets=['u1'])

        # Third Phase (burn)
        burn2 = dm.Phase(ode_class=FiniteBurnODE,
                         transcription=dm.GaussLobatto(num_segments=5, order=3, compressed=False))

        traj.add_phase('coast', coast)
        traj.add_phase('burn2', burn2)

        burn2.set_time_options(initial_bounds=(0.5, 50), duration_bounds=(.5, 10), initial_ref=10,
                               units='TU')
        burn2.add_state('r', fix_initial=False, fix_final=True, defect_scaler=100.0,
                        rate_source='r_dot', units='DU')
        burn2.add_state('theta', fix_initial=False, fix_final=False, defect_scaler=100.0,
                        rate_source='theta_dot', units='rad')
        burn2.add_state('vr', fix_initial=False, fix_final=True, defect_scaler=1000.0,
                        rate_source='vr_dot', units='DU/TU')
        burn2.add_state('vt', fix_initial=False, fix_final=True, defect_scaler=1000.0,
                        rate_source='vt_dot', units='DU/TU')
        burn2.add_state('accel', fix_initial=False, fix_final=False, defect_scaler=1.0,
                        rate_source='at_dot', units='DU/TU**2')
        burn2.add_state('deltav', fix_initial=False, fix_final=False, defect_scaler=1.0,
                        rate_source='deltav_dot', units='DU/TU')

        burn2.add_objective('deltav', loc='final', scaler=100.0)

        burn2.add_control('u1', rate_continuity=True, rate2_continuity=True, units='deg',
                          scaler=0.01, lower=-90, upper=90)

        burn1.add_timeseries_output('pos_x')
        coast.add_timeseries_output('pos_x')
        burn2.add_timeseries_output('pos_x')

        burn1.add_timeseries_output('pos_y')
        coast.add_timeseries_output('pos_y')
        burn2.add_timeseries_output('pos_y')

        # Link Phases
        traj.link_phases(phases=['burn1', 'coast', 'burn2'],
                         vars=['time', 'r', 'theta', 'vr', 'vt', 'deltav'])

        traj.link_phases(phases=['burn1', 'burn2'], vars=['accel'])

        p.model.add_subsystem('traj', subsys=traj)

        # Finish Problem Setup

        # Needed to move the direct solver down into the phases for use with MPI.
        #  - After moving down, used fewer iterations (about 30 less)

        p.driver.add_recorder(om.SqliteRecorder('two_burn_orbit_raise_example.db'))

        p.setup(check=True, mode='fwd')

        # Set Initial Guesses
        p.set_val('traj.parameters:c', val=1.5, units='DU/TU')

        burn1 = p.model.traj.phases.burn1
        burn2 = p.model.traj.phases.burn2
        coast = p.model.traj.phases.coast

        p.set_val('traj.burn1.t_initial', val=0.0)
        p.set_val('traj.burn1.t_duration', val=2.25)
        p.set_val('traj.burn1.states:r', val=burn1.interp('r', [1, 1.5]))
        p.set_val('traj.burn1.states:theta', val=burn1.interp('theta', [0, 1.7]))
        p.set_val('traj.burn1.states:vr', val=burn1.interp('vr', [0, 0]))
        p.set_val('traj.burn1.states:vt', val=burn1.interp('vt', [1, 1]))
        p.set_val('traj.burn1.states:accel', val=burn1.interp('accel', [0.1, 0]))
        p.set_val('traj.burn1.states:deltav', val=burn1.interp('deltav', [0, 0.1]))
        p.set_val('traj.burn1.controls:u1', val=burn1.interp('u1', [-3.5, 13.0]))

        p.set_val('traj.coast.t_initial', val=2.25)
        p.set_val('traj.coast.t_duration', val=3.0)

        p.set_val('traj.coast.states:r', val=coast.interp('r', [1.3, 1.5]))
        p.set_val('traj.coast.states:theta', val=coast.interp('theta', [2.1767, 1.7]))
        p.set_val('traj.coast.states:vr', val=coast.interp('vr', [0.3285, 0]))
        p.set_val('traj.coast.states:vt', val=coast.interp('vt', [0.97, 1]))
        p.set_val('traj.coast.states:accel', val=coast.interp('accel', [0, 0]))

        p.set_val('traj.burn2.t_initial', val=5.25)
        p.set_val('traj.burn2.t_duration', val=1.75)

        p.set_val('traj.burn2.states:r', val=burn2.interp('r', [1, 3.]))
        p.set_val('traj.burn2.states:theta', val=burn2.interp('theta', [0, 4.0]))
        p.set_val('traj.burn2.states:vr', val=burn2.interp('vr', [0, 0]))
        p.set_val('traj.burn2.states:vt', val=burn2.interp('vt', [1, np.sqrt(1 / 3.)]))
        p.set_val('traj.burn2.states:deltav', val=burn2.interp('deltav', [0.1, 0.2]))
        p.set_val('traj.burn2.states:accel', val=burn2.interp('accel', [0.1, 0]))

        p.set_val('traj.burn2.controls:u1', val=burn2.interp('u1', [0, 0]))

        dm.run_problem(p)

        assert_near_equal(p.get_val('traj.burn2.states:deltav')[-1], 0.3995,
                          tolerance=2.0E-3)

        #
        # Plot results
        #
        traj = p.model.traj
        exp_out = traj.simulate()

        fig = plt.figure(figsize=(8, 4))
        fig.suptitle('Two Burn Orbit Raise Solution')
        ax_u1 = plt.subplot2grid((2, 2), (0, 0))
        ax_deltav = plt.subplot2grid((2, 2), (1, 0))
        ax_xy = plt.subplot2grid((2, 2), (0, 1), rowspan=2)

        span = np.linspace(0, 2 * np.pi, 100)
        ax_xy.plot(np.cos(span), np.sin(span), 'k--', lw=1)
        ax_xy.plot(3 * np.cos(span), 3 * np.sin(span), 'k--', lw=1)
        ax_xy.set_xlim(-4.5, 4.5)
        ax_xy.set_ylim(-4.5, 4.5)

        ax_xy.set_xlabel('x ($R_e$)')
        ax_xy.set_ylabel('y ($R_e$)')

        ax_u1.set_xlabel('time ($TU$)')
        ax_u1.set_ylabel('$u_1$ ($deg$)')
        ax_u1.grid(True)

        ax_deltav.set_xlabel('time ($TU$)')
        ax_deltav.set_ylabel('${\Delta}v$ ($DU/TU$)')
        ax_deltav.grid(True)

        t_sol = dict((phs, p.get_val('traj.{0}.timeseries.time'.format(phs)))
                     for phs in ['burn1', 'coast', 'burn2'])
        x_sol = dict((phs, p.get_val('traj.{0}.timeseries.pos_x'.format(phs)))
                     for phs in ['burn1', 'coast', 'burn2'])
        y_sol = dict((phs, p.get_val('traj.{0}.timeseries.pos_y'.format(phs)))
                     for phs in ['burn1', 'coast', 'burn2'])
        dv_sol = dict((phs, p.get_val('traj.{0}.timeseries.states:deltav'.format(phs)))
                      for phs in ['burn1', 'coast', 'burn2'])
        u1_sol = dict((phs, p.get_val('traj.{0}.timeseries.controls:u1'.format(phs), units='deg'))
                      for phs in ['burn1', 'burn2'])

        t_exp = dict((phs, exp_out.get_val('traj.{0}.timeseries.time'.format(phs)))
                     for phs in ['burn1', 'coast', 'burn2'])
        x_exp = dict((phs, exp_out.get_val('traj.{0}.timeseries.pos_x'.format(phs)))
                     for phs in ['burn1', 'coast', 'burn2'])
        y_exp = dict((phs, exp_out.get_val('traj.{0}.timeseries.pos_y'.format(phs)))
                     for phs in ['burn1', 'coast', 'burn2'])
        dv_exp = dict((phs, exp_out.get_val('traj.{0}.timeseries.states:deltav'.format(phs)))
                      for phs in ['burn1', 'coast', 'burn2'])
        u1_exp = dict((phs, exp_out.get_val('traj.{0}.timeseries.controls:u1'.format(phs),
                                            units='deg'))
                      for phs in ['burn1', 'burn2'])

        for phs in ['burn1', 'coast', 'burn2']:
            try:
                ax_u1.plot(t_exp[phs], u1_exp[phs], '-', marker=None, color='C0')
                ax_u1.plot(t_sol[phs], u1_sol[phs], 'o', mfc='C1', mec='C1', ms=3)
            except KeyError:
                pass

            ax_deltav.plot(t_exp[phs], dv_exp[phs], '-', marker=None, color='C0')
            ax_deltav.plot(t_sol[phs], dv_sol[phs], 'o', mfc='C1', mec='C1', ms=3)

            ax_xy.plot(x_exp[phs], y_exp[phs], '-', marker=None, color='C0', label='explicit')
            ax_xy.plot(x_sol[phs], y_sol[phs], 'o', mfc='C1', mec='C1', ms=3, label='implicit')

        plt.show()


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



from fnmatch import filter
import warnings

import numpy as np

import openmdao.api as om

from ..transcription_base import TranscriptionBase
from .components import SegmentSimulationComp, SegmentStateMuxComp, \
    SolveIVPControlGroup, SolveIVPPolynomialControlGroup, SolveIVPTimeseriesOutputComp
from ..common import TimeComp
from ...utils.misc import get_rate_units
from ...utils.introspection import get_promoted_vars, get_targets, get_source_metadata, get_target_metadata
from ...utils.indexing import get_src_indices_by_row


class SolveIVP(TranscriptionBase):
    """
    The SolveIVP Transcription class.

    SolveIVP transcription in Dymos uses the scipy.simulate.solve_ivp method to explicitly integrate
    the states from the phase initial time to the phase final time.

    SolveIVP transcription does not currently support optimization since it does not propagate
    analytic derivatives through the ODE.

    Parameters
    ----------
    grid_data : GridData
        Grid data for this phases.
    **kwargs : dict
        Dictionary of optional arguments.
    """
    def __init__(self, grid_data=None, **kwargs):
        super(SolveIVP, self).__init__(**kwargs)
        self.grid_data = grid_data
        self._rhs_source = 'ode'

    def initialize(self):
        """
        Declare transcription options.
        """
        super(SolveIVP, self).initialize()

        self.options.declare('output_nodes_per_seg', default=None, types=(int,), allow_none=True,
                             desc='If None, results are provided at the all nodes within each'
                                  'segment.  If an int (n) then results are provided at n '
                                  'equally distributed points in time within each segment.')

        self.options.declare('reports', default=False, desc='Reports setting for the subproblem.')

    def init_grid(self):
        """
        Setup the GridData object for the Transcription.
        """
        pass

    def setup_time(self, phase):
        """
        Setup the time component.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        time_options = phase.time_options
        time_units = time_options['units']
        num_seg = self.grid_data.num_segments
        grid_data = self.grid_data
        output_nodes_per_seg = self.options['output_nodes_per_seg']

        super(SolveIVP, self).setup_time(phase)

        if output_nodes_per_seg is None:
            # Case 1:  Compute times at 'all' node set.
            num_nodes = grid_data.num_nodes
            node_ptau = grid_data.node_ptau
            node_dptau_dstau = grid_data.node_dptau_dstau
        else:
            # Case 2:  Compute times at n equally distributed points per segment.
            num_nodes = num_seg * output_nodes_per_seg
            node_stau = np.linspace(-1, 1, output_nodes_per_seg)
            node_ptau = np.empty(0, )
            node_dptau_dstau = np.empty(0, )
            # Append our nodes in phase tau space
            for iseg in range(num_seg):
                v0 = grid_data.segment_ends[iseg]
                v1 = grid_data.segment_ends[iseg + 1]
                node_ptau = np.concatenate((node_ptau, v0 + 0.5 * (node_stau + 1) * (v1 - v0)))
                node_dptau_dstau = np.concatenate((node_dptau_dstau,
                                                   0.5 * (v1 - v0) * np.ones_like(node_stau)))

        time_comp = TimeComp(num_nodes=num_nodes, node_ptau=node_ptau,
                             node_dptau_dstau=node_dptau_dstau, units=time_units)

        phase.add_subsystem('time', time_comp, promotes=['*'])

    def configure_time(self, phase):
        """
        Configure the inputs/outputs on the time component.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        super(SolveIVP, self).configure_time(phase)
        num_seg = self.grid_data.num_segments
        grid_data = self.grid_data
        output_nodes_per_seg = self.options['output_nodes_per_seg']
        ode = self._get_ode(phase)

        phase.time.configure_io()

        for i in range(num_seg):
            if output_nodes_per_seg is None:
                i1, i2 = grid_data.subset_segment_indices['all'][i, :]
                src_idxs = grid_data.subset_node_indices['all'][i1:i2]
            else:
                src_idxs = np.arange(i * output_nodes_per_seg, output_nodes_per_seg * (i + 1),
                                     dtype=int)
            phase.connect('time', f'segment_{i}.time', src_indices=src_idxs, flat_src_indices=True)
            phase.connect('time_phase', f'segment_{i}.time_phase', src_indices=src_idxs,
                          flat_src_indices=True)

            phase.segments.promotes(f'segment_{i}', inputs=['t_initial', 't_duration'])

        options = phase.time_options

        # The tuples here are (name, user_specified_targets, dynamic)
        for name, targets, dynamic in [('time', options['targets'], True),
                                       ('time_phase', options['time_phase_targets'], True)]:
            if targets:
                phase.connect(name, [f'ode.{t}' for t in targets])

        for name, targets, dynamic in [('t_initial', options['t_initial_targets'], False),
                                       ('t_duration', options['t_duration_targets'], False)]:

            shape, units, static_target = get_target_metadata(ode, name=name,
                                                              user_targets=targets,
                                                              user_units=options['units'],
                                                              user_shape=(1,))
            if shape == (1,):
                src_idxs = None
                flat_src_idxs = None
                src_shape = None
            else:
                src_idxs = np.zeros(self.grid_data.subset_num_nodes['all'])
                flat_src_idxs = True
                src_shape = (1,)

            for t in targets:
                phase.promotes('ode', inputs=[(t, name)], src_indices=src_idxs,
                               flat_src_indices=flat_src_idxs, src_shape=src_shape)
            if targets:
                phase.set_input_defaults(name=name,
                                         val=np.ones((1,)),
                                         units=options['units'])

        phase.connect('dt_dstau', 'timeseries.dt_dstau')

    def setup_states(self, phase):
        """
        Setup the states for this transcription.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        phase.add_subsystem('indep_states', om.IndepVarComp(),
                            promotes_outputs=['*'])

    def configure_states(self, phase):
        """
        Configure state connections post-introspection.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        super().configure_states(phase)
        num_seg = self.grid_data.num_segments

        for state_name, options in phase.state_options.items():
            phase.indep_states.add_output(f'initial_states:{state_name}',
                                          val=np.ones(((1,) + options['shape'])),
                                          units=options['units'])

        for state_name, options in phase.state_options.items():
            # Connect the initial state to the first segment
            src_idxs = get_src_indices_by_row([0], options['shape'])

            phase.connect(f'initial_states:{state_name}',
                          f'segment_0.initial_states:{state_name}',
                          src_indices=(src_idxs,), flat_src_indices=True)

            phase.connect(f'segment_0.states:{state_name}',
                          f'state_mux_comp.segment_0_states:{state_name}')

            if options['targets']:
                phase.connect(f'state_mux_comp.states:{state_name}',
                              [f'ode.{t}' for t in options['targets']])

            # Connect the final state in segment n to the initial state in segment n + 1
            for i in range(1, num_seg):
                if self.options['output_nodes_per_seg'] is None:
                    nnps_i = self.grid_data.subset_num_nodes_per_segment['all'][i]
                else:
                    nnps_i = self.options['output_nodes_per_seg']

                src_idxs = get_src_indices_by_row([nnps_i-1], shape=options['shape'])
                phase.connect(f'segment_{i - 1}.states:{state_name}',
                              f'segment_{i}.initial_states:{state_name}',
                              src_indices=(src_idxs,), flat_src_indices=True)

                phase.connect(f'segment_{i}.states:{state_name}',
                              f'state_mux_comp.segment_{i}_states:{state_name}')

    def setup_ode(self, phase):
        """
        Setup the ode for this transcription.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        gd = self.grid_data
        num_seg = gd.num_segments

        segments_group = phase.add_subsystem(name='segments', subsys=om.Group(),
                                             promotes_outputs=['*'], promotes_inputs=['*'])

        for i in range(num_seg):
            seg_i_comp = SegmentSimulationComp(
                index=i,
                simulate_options=phase.simulate_options,
                grid_data=self.grid_data,
                ode_class=phase.options['ode_class'],
                ode_init_kwargs=phase.options['ode_init_kwargs'],
                time_options=phase.time_options,
                state_options=phase.state_options,
                control_options=phase.control_options,
                polynomial_control_options=phase.polynomial_control_options,
                parameter_options=phase.parameter_options,
                output_nodes_per_seg=self.options['output_nodes_per_seg'],
                reports=self.options['reports'])

            segments_group.add_subsystem(f'segment_{i}', subsys=seg_i_comp)

        # scipy.integrate.solve_ivp does not actually evaluate the ODE at the desired output points,
        # but just returns the time and interpolated integrated state values there instead. We need
        # to instantiate a second ODE group that will call the ODE at those points so that we can
        # accurately obtain timeseries for ODE outputs.
        phase.add_subsystem('state_mux_comp',
                            SegmentStateMuxComp(grid_data=gd, state_options=phase.state_options,
                                                output_nodes_per_seg=self.options['output_nodes_per_seg']))

        if self.options['output_nodes_per_seg'] is None:
            self.num_output_nodes = gd.subset_num_nodes['all']
        else:
            self.num_output_nodes = num_seg * self.options['output_nodes_per_seg']

        phase.add_subsystem('ode', phase.options['ode_class'](num_nodes=self.num_output_nodes,
                                                              **phase.options['ode_init_kwargs']))

    def configure_ode(self, phase):
        """
        Create connections to the introspected states.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        gd = self.grid_data
        num_seg = gd.num_segments

        for i in range(num_seg):
            seg_comp = phase.segments._get_subsystem(f'segment_{i}')
            seg_comp.configure_io()

    def setup_controls(self, phase):
        """
        Setup the control group.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        output_nodes_per_seg = self.options['output_nodes_per_seg']

        phase._check_control_options()

        if phase.control_options:
            control_group = SolveIVPControlGroup(control_options=phase.control_options,
                                                 time_units=phase.time_options['units'],
                                                 grid_data=self.grid_data,
                                                 output_nodes_per_seg=output_nodes_per_seg)

            phase.add_subsystem('control_group',
                                subsys=control_group,
                                promotes=['controls:*', 'control_values:*', 'control_values_all:*',
                                          'control_rates:*'])

    def configure_controls(self, phase):
        """
        Configure the inputs/outputs for the controls.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        grid_data = self.grid_data

        if phase.control_options:
            phase.control_group.configure_io()
            phase.connect('dt_dstau', 'control_group.dt_dstau')

        for name, options in phase.control_options.items():
            for i in range(grid_data.num_segments):
                i1, i2 = grid_data.subset_segment_indices['control_disc'][i, :]
                seg_idxs = grid_data.subset_node_indices['control_disc'][i1:i2]
                src_idxs = get_src_indices_by_row(row_idxs=seg_idxs, shape=options['shape'])
                phase.connect(src_name=f'control_values_all:{name}',
                              tgt_name=f'segment_{i}.controls:{name}',
                              src_indices=(src_idxs,), flat_src_indices=True)

            if options['targets']:
                phase.connect(f'control_values:{name}', [f'ode.{t}' for t in options['targets']])

            if options['rate_targets']:
                phase.connect(f'control_rates:{name}_rate', [f'ode.{t}' for t in options['rate_targets']])

            if options['rate2_targets']:
                phase.connect(f'control_rates:{name}_rate2', [f'ode.{t}' for t in options['rate2_targets']])

    def setup_polynomial_controls(self, phase):
        """
        Adds the polynomial control group to the model if any polynomial controls are present.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        if phase.polynomial_control_options:
            sys = SolveIVPPolynomialControlGroup(grid_data=self.grid_data,
                                                 polynomial_control_options=phase.polynomial_control_options,
                                                 time_units=phase.time_options['units'],
                                                 output_nodes_per_seg=self.options['output_nodes_per_seg'])
            phase.add_subsystem('polynomial_control_group', subsys=sys,
                                promotes_inputs=['*'], promotes_outputs=['*'])

    def configure_polynomial_controls(self, phase):
        """
        Configure the inputs/outputs for the polynomial controls.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        # In transcription_base, we get the control units/shape from the target, and then call
        # configure on the control_group.
        super(SolveIVP, self).configure_polynomial_controls(phase)

        # Additional connections.
        for name, options in phase.polynomial_control_options.items():
            targets = options['targets']

            for iseg in range(self.grid_data.num_segments):
                phase.connect(src_name=f'polynomial_controls:{name}',
                              tgt_name=f'segment_{iseg}.polynomial_controls:{name}')

            if options['targets']:
                phase.connect(f'polynomial_control_values:{name}', [f'ode.{t}' for t in targets])

            if options['rate_targets']:
                phase.connect(f'polynomial_control_rates:{name}_rate',
                              [f'ode.{t}' for t in targets])

            if options['rate2_targets']:
                phase.connect(f'polynomial_control_rates:{name}_rate2',
                              [f'ode.{t}' for t in targets])

    def configure_parameters(self, phase):
        """
        Configure parameter promotion.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        super(SolveIVP, self).configure_parameters(phase)

        gd = self.grid_data

        # We also need to take care of the segments.
        segs = phase._get_subsystem('segments')

        for name, options in phase.parameter_options.items():
            # prom_name = f'parameters:{name}'
            shape = options['shape']
            units = options['units']

            for i in range(gd.num_segments):
                seg_comp = segs._get_subsystem(f'segment_{i}')
                seg_comp.add_input(name=f'parameters:{name}', val=np.ones(shape), units=units,
                                   desc=f'values of parameter {name}.')
                # phase.connect(f'parameter_vals:{name}', f'segment_{i}.{prom_name}')

    def setup_defects(self, phase):
        """
        Not used in SolveIVP.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        pass

    def configure_defects(self, phase):
        """
        Not used in SolveIVP.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        pass

    def configure_objective(self, phase):
        """
        Not used in SolveIVP.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        pass

    def configure_path_constraints(self, phase):
        """
        Not used in SolveIVP.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        pass

    def _configure_boundary_constraints(self, phase):
        """
        Not used in SolveIVP.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        pass

    def setup_solvers(self, phase):
        """
        Not used in SolveIVP.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        pass

    def configure_solvers(self, phase):
        """
        Not used in SolveIVP.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        pass

    def setup_timeseries_outputs(self, phase):
        """
        Setup the timeseries for this transcription.

        Parameters
        ----------
        phase : dymos.Phase
            The phase object to which this transcription instance applies.
        """
        gd = self.grid_data

        timeseries_comp = \
            SolveIVPTimeseriesOutputComp(input_grid_data=gd,
                                         output_nodes_per_seg=self.options['output_nodes_per_seg'],
                                         time_units=phase.time_options['units'])

        phase.add_subsystem('timeseries', subsys=timeseries_comp)

        # Remove all subsequent timeseries
        for ts_name in list(phase._timeseries.keys()):
            if ts_name != 'timeseries':
                phase._timeseries.pop(ts_name)

    def get_parameter_connections(self, name, phase):
        """
        Returns info about a parameter's target connections in the phase.

        Parameters
        ----------
        name : str
            The name of the parameter for which connection information is desired.
        phase : dymos.Phase
            The phase object to which this transcription applies.

        Returns
        -------
        list of (paths, indices)
            A list containing a tuple of target paths and corresponding src_indices to which the
            given design variable is to be connected.
        """
        connection_info = []
        num_seg = self.grid_data.num_segments
        output_nodes_per_seg = self.options['output_nodes_per_seg']
        num_final_ode_nodes = self.grid_data.subset_num_nodes['all'] \
            if output_nodes_per_seg is None else num_seg * output_nodes_per_seg

        if name in phase.parameter_options:
            options = phase.parameter_options[name]

            static = options['static_target']
            shape = options['shape']

            # Get connections to each segment
            gd = self.grid_data

            for i in range(gd.num_segments):
                phase.connect(f'parameter_vals:{name}', f'segment_{i}.parameters:{name}')

            # Connections to the final ODE
            ode_tgts = get_targets(ode=phase.ode, name=name, user_targets=options['targets'])
            if not static:
                src_idxs_raw = np.zeros(num_final_ode_nodes, dtype=int)
                src_idxs = get_src_indices_by_row(src_idxs_raw, shape)
                if shape == (1,):
                    src_idxs = src_idxs.ravel()
            else:
                src_idxs_raw = np.zeros(1, dtype=int)
                src_idxs = get_src_indices_by_row(src_idxs_raw, shape)
                src_idxs = np.squeeze(src_idxs, axis=0)

            connection_info.append(([f'ode.{tgt}' for tgt in ode_tgts], (src_idxs,)))

        return connection_info

    def _get_timeseries_var_source(self, var, output_name, phase):
        """
        Return the source path and indices for a given variable to be connected to a timeseries.

        Parameters
        ----------
        var : str
            Name of the timeseries variable whose source is desired.
        output_name : str
            Name of the timeseries output whose source is desired.
        phase : dymos.Phase
            Phase object containing the variable, either as state, time, control, etc., or as an ODE output.

        Returns
        -------
        meta : dict
            Metadata pertaining to the variable at the given path. This dict contains 'src' (the path to the
            timeseries source), 'src_idxs' (an array of the
            source indices), 'units' (the units of the source variable), and 'shape' (the shape of the variable at
            a given node).
        """
        gd = self.grid_data
        var_type = phase.classify_var(var)
        time_units = phase.time_options['units']

        transcription = phase.options['transcription']
        ode = transcription._get_ode(phase)
        ode_outputs = get_promoted_vars(ode, 'output')

        # The default for node_idxs, applies to everything except states and parameters.
        node_idxs = None
        meta = {}

        # Determine the path to the variable
        if var_type == 'time':
            path = 'time'
            src_units = time_units
            src_shape = (1,)
        elif var_type == 'time_phase':
            path = 'time_phase'
            src_units = time_units
            src_shape = (1,)
        elif var_type == 'state':
            path = f'state_mux_comp.states:{var}'
            src_units = phase.state_options[var]['units']
            src_shape = phase.state_options[var]['shape']
        elif var_type in ['indep_control', 'input_control']:
            path = f'control_values:{var}'
            src_units = phase.control_options[var]['units']
            src_shape = phase.control_options[var]['shape']
        elif var_type == 'control_rate':
            control_name = var[:-5]
            path = f'control_rates:{control_name}_rate'
            control_name = var[:-5]
            src_units = get_rate_units(phase.control_options[control_name]['units'], time_units, deriv=1)
            src_shape = phase.control_options[control_name]['shape']
        elif var_type == 'control_rate2':
            control_name = var[:-6]
            path = f'control_rates:{control_name}_rate2'
            src_units = get_rate_units(phase.control_options[control_name]['units'], time_units, deriv=2)
            src_shape = phase.control_options[control_name]['shape']
        elif var_type in ['indep_polynomial_control', 'input_polynomial_control']:
            path = f'polynomial_control_values:{var}'
            src_units = phase.polynomial_control_options[var]['units']
            src_shape = phase.polynomial_control_options[var]['shape']
        elif var_type == 'polynomial_control_rate':
            control_name = var[:-5]
            path = f'polynomial_control_rates:{control_name}_rate'
            control = phase.polynomial_control_options[control_name]
            src_units = get_rate_units(control['units'], time_units, deriv=1)
            src_shape = control['shape']
        elif var_type == 'polynomial_control_rate2':
            control_name = var[:-6]
            path = f'polynomial_control_rates:{control_name}_rate2'
            control = phase.polynomial_control_options[control_name]
            src_units = get_rate_units(control['units'], time_units, deriv=1)
            src_shape = control['shape']
        elif var_type == 'parameter':
            path = f'parameter_vals:{var}'
            num_seg = self.grid_data.num_segments
            node_idxs = np.zeros(num_seg * self.options['output_nodes_per_seg'], dtype=int)
            src_units = phase.parameter_options[var]['units']
            src_shape = phase.parameter_options[var]['shape']
        else:
            # Failed to find variable, assume it is in the ODE
            path = f'ode.{var}'
            meta = get_source_metadata(ode_outputs, src=var)
            src_shape = meta['shape']
            src_units = meta['units']
            src_tags = meta['tags']
            if 'dymos.static_output' in src_tags:
                raise RuntimeError(f'ODE output {var} is tagged with "dymos.static_output" and cannot be a '
                                   f'timeseries output.')

        src_idxs = None if node_idxs is None else om.slicer[node_idxs, ...]

        meta['src'] = path
        meta['src_idxs'] = src_idxs
        meta['units'] = src_units
        meta['shape'] = src_shape

        return meta



import copy

import numpy as np

from ...utils.lgr import lgr
from ...utils.lgl import lgl
from ...utils.interpolate import LagrangeBarycentricInterpolant
from ..error_estimation import interpolation_lagrange_matrix, eval_ode_on_grid


def split_segments(old_seg_ends, B):
    """
    Compute the new segment ends for the refined grid by splitting the necessary segments.

    Parameters
    ----------
    old_seg_ends : ndarray
        Segment ends of grid on which the problem was solved.
    B : ndrray of int
        Number of segments to be split into.

    Returns
    -------
    ndarray
        Segment ends of refined grid.
    """
    new_segment_ends = []
    for q in range(0, B.size):
        new_ends = np.linspace(old_seg_ends[q], old_seg_ends[q + 1], B[q] + 1)
        new_segment_ends.extend(new_ends[:-1])
    new_segment_ends.append(1)
    return np.asarray(new_segment_ends)


def merge_segments(old_seg_ends, seg_merge):
    """
    Compute the new segment ends for the refined grid by merging the unnecessary segments.

    Parameters
    ----------
    old_seg_ends : ndarray
        Segment ends of grid on which the problem was solved.
    seg_merge : ndarray of bool
        True for segments that are to be merged.

    Returns
    -------
    ndarray
        Segment ends of refined grid.
    """
    new_seg_ends = [old_seg_ends[0]]
    for q in range(1, seg_merge.size):
        if seg_merge[q]:
            continue
        new_seg_ends.append(old_seg_ends[q])
    new_seg_ends.append(1)
    return np.asarray(new_seg_ends)


class HPAdaptive:
    """
    Grid refinement object for the hp grid refinement algorithm.

    The error on a solved phase is evaluated. If error exceeds chosen tolerance, the grid is refined following the
    p-then-h refinement algorithm.
    Patterson, M. A., Hager, W. W., and Rao. A. V., “Adaptive mesh refinement method for optimal control using
    non-smoothness detection and mesh size reduction”, Journal of the Franklin Institute 352 (2015) 4081–4106

    Parameters
    ----------
    phases : Phase
        The Phase object representing the solved phase.
    """

    def __init__(self, phases):
        self.phases = phases
        self.error = {}
        self.iteration_number = 0
        self.previous_error = {}
        self.previous_gd = {}
        self.previous_x_dd = {}
        self.parent_seg_map = {}

    def refine_first_iter(self, refine_results):
        """
        Refine the grid during the first iteration.

        Compute the order, number of nodes, and segment ends required for the new grid
        and assigns them to the transcription of each phase. Method of refinement is
        different for the first iteration and is done separately

        Parameters
        ----------
        refine_results : dict
            A dictionary where each key is the path to a phase in the problem, and the
            associated value are various properties of that phase needed by the refinement
            algorithm.  refine_results is returned by check_error.  This method modifies it
            in place, adding the new_num_segments, new_order, and new_segment_ends.

        Returns
        -------
        dict
            A dictionary of phase paths : phases which were refined.
        """

        for phase_path, phase_refinement_results in refine_results.items():
            phase = self.phases[phase_path]
            tx = phase.options['transcription']
            gd = tx.grid_data
            refine_tol = phase.refine_options['tolerance']
            refine_min_order = phase.refine_options['min_order']
            self.previous_gd[phase_path] = gd
            self.previous_x_dd[phase_path] = {}
            self.error[phase_path] = refine_results[phase_path]['max_rel_error']

            # Get information about current grid
            num_scalar_states = 0
            for state_name, options in phase.state_options.items():
                shape = options['shape']
                size = np.prod(shape)
                num_scalar_states += size

            seg_order = gd.transcription_order
            seg_ends = gd.segment_ends

            need_refine = phase_refinement_results['need_refinement']
            if not phase.refine_options['refine'] or not np.any(need_refine):
                refine_results[phase_path]['new_order'] = seg_order
                refine_results[phase_path]['new_num_segments'] = gd.num_segments
                refine_results[phase_path]['new_segment_ends'] = seg_ends
                continue

            left_end_idxs = gd.subset_node_indices['segment_ends'][0::2]
            left_end_idxs = np.append(left_end_idxs, gd.subset_num_nodes['all'])

            # obtain state and state rate histories from timeseries output
            L, D = interpolation_lagrange_matrix(gd, gd)
            x, _, _, x_d = eval_ode_on_grid(phase=phase, transcription=tx)

            # create and store second derivative information using differentiation matrix
            for state_name, options in phase.state_options.items():
                self.previous_x_dd[phase_path][state_name] = D @ x_d[state_name]

            gd = phase.options['transcription'].grid_data
            numseg = gd.num_segments

            merge_seg = np.zeros(numseg, dtype=bool)

            # In the first iteration no segments are split. All segments not meeting error requirements have their order
            # increased by 3 for Radau transcription and by 2 for GL
            inc_seg_order_idxs = np.where(need_refine)
            P = np.zeros(numseg)
            if gd.transcription == 'radau-ps':
                P[inc_seg_order_idxs] = 3
            elif gd.transcription == 'gauss-lobatto':
                P[inc_seg_order_idxs] = 2  # GL transcription does not allow even segment orders
            new_order = (seg_order + P).astype(int)

            h = 0.5 * (seg_ends[1:] - seg_ends[:-1])

            # Segments which should be checked for combining
            # Only segments where adjacent segments are also below error tolerance may be combined
            # Segments must be same order
            check_comb_indx = np.where(np.logical_and(np.logical_and(np.logical_and(np.invert(need_refine[:-1]),
                                                                                    np.invert(need_refine[1:])),
                                                                     new_order[:-1] == new_order[1:]),
                                                      new_order[:-1] == refine_min_order))[0]

            # segments under error tolerance but may not be combined are checked to have their order reduced
            reduce_order_indx = np.setdiff1d(np.where(np.invert(need_refine)), check_comb_indx)

            # reduce segment order where error is much below the tolerance
            # Order reduction is done by creating a power series representation of the state data on the segment
            # Series order is progressively reduced until removal of additional terms would lead to error > tolerance
            if reduce_order_indx.size > 0:
                # compute normalization factor beta
                beta = {}
                for state_name, options in phase.state_options.items():
                    beta[state_name] = 0
                    for k in range(0, numseg):
                        beta_seg = np.max(np.abs(x[state_name][left_end_idxs[k]:left_end_idxs[k + 1]]))
                        if beta_seg > beta[state_name]:
                            beta[state_name] = beta_seg
                    beta[state_name] += 1

                for k in np.nditer(reduce_order_indx):
                    seg_size = {'radau-ps': seg_order[k] + 1, 'gauss-lobatto': seg_order[k]}
                    if seg_order[k] == refine_min_order:
                        continue
                    new_order_state = {}
                    new_order[k] = seg_order[k]
                    a = np.zeros((seg_size[gd.transcription], seg_size[gd.transcription]))
                    s, _ = lgr(seg_order[k], include_endpoint=True)
                    if gd.transcription == 'gauss-lobatto':
                        s, _ = lgl(seg_order[k])
                    for j in range(0, seg_size[gd.transcription]):
                        roots = s[s != s[j]]
                        Q = np.poly(roots)
                        a[:, j] = Q / np.polyval(Q, s[j])

                    for state_name, options in phase.state_options.items():
                        new_order_state[state_name] = seg_order[k]
                        b = a @ x[state_name][left_end_idxs[k]:left_end_idxs[k + 1]]
                        for i in range(seg_size[gd.transcription] - 1, phase.refine_options['min_order'], -1):
                            state_error = np.abs(b[i]) / beta[state_name]
                            if state_error < refine_tol:
                                if gd.transcription == 'gauss-lobatto':
                                    if new_order_state[state_name] < i - 2 <= refine_min_order:
                                        new_order_state[state_name] = i - 2
                                elif gd.transcription == 'radau-ps':
                                    if new_order_state[state_name] < i - 1 <= refine_min_order:
                                        new_order_state[state_name] = i - 1
                        new_order[k] = max(new_order_state.values())

            # combine unnecessary segments
            # The first of the two segments is extrapolated onto the second segment
            # The extrapolation is checked against the current solution
            # If they match closely enough the segments may be merged
            if check_comb_indx.size > 0:
                for k in np.nditer(check_comb_indx):
                    seg_size = {'radau-ps': seg_order[k] + 1, 'gauss-lobatto': seg_order[k]}
                    if merge_seg[k]:
                        continue
                    a = np.zeros((seg_size[gd.transcription], seg_size[gd.transcription]))
                    h_ = np.maximum(h[k], h[k + 1])
                    s, _ = lgr(new_order[k].astype(int), include_endpoint=True)
                    if gd.transcription == 'gauss-lobatto':
                        s, _ = lgl(new_order[k])
                    for j in range(0, seg_size[gd.transcription]):
                        roots = s[s != s[j]]
                        Q = np.poly(roots)
                        a[:, j] = Q / np.polyval(Q, s[j])

                    merge_seg[k + 1] = True

                    for state_name, options in phase.state_options.items():
                        beta = 1 + np.max(x[state_name][left_end_idxs[k]:left_end_idxs[k + 1]])
                        c = a @ x[state_name][left_end_idxs[k]:left_end_idxs[k + 1]]
                        b = np.multiply(c.ravel(), np.array([(h_ / h[k]) ** l for l in
                                                             range(seg_size[gd.transcription])]))
                        b_hat = np.multiply(c.ravel(),
                                            np.array([(h_ / h[k + 1]) ** l for l in range(seg_size[gd.transcription])]))
                        err_val = np.dot(np.absolute(b - b_hat).ravel(),
                                         np.array([2 ** l for l in range(seg_size[gd.transcription])])) / beta

                        if err_val > phase.refine_options['tolerance'] and merge_seg[k + 1]:
                            merge_seg[k + 1] = False

            new_segment_ends = merge_segments(gd.segment_ends, merge_seg)
            new_num_segments = new_segment_ends.shape[0] - 1
            new_order = np.delete(new_order, np.where(merge_seg), axis=None)

            # Create a dictionary with the information on which segment included the time of each segment
            self.parent_seg_map[phase_path] = np.zeros(new_num_segments, dtype=int)
            for i in range(1, new_num_segments):
                for j in range(1, numseg):
                    if new_segment_ends[i] == gd.segment_ends[j]:
                        self.parent_seg_map[phase_path][i] = int(j)
                        break
                    elif gd.segment_ends[j - 1] < new_segment_ends[i] < gd.segment_ends[j]:
                        self.parent_seg_map[phase_path][i] = int(j - 1)
                        break

            if gd.transcription == 'gauss-lobatto':
                new_order[(new_order % 2) == 0] = new_order[(new_order % 2) == 0] + 1

            refine_results[phase_path]['new_order'] = new_order
            refine_results[phase_path]['new_num_segments'] = new_num_segments
            refine_results[phase_path]['new_segment_ends'] = new_segment_ends

            tx.options['order'] = new_order
            tx.options['num_segments'] = new_num_segments
            tx.options['segment_ends'] = new_segment_ends
            tx.init_grid()
            self.previous_error[phase_path] = self.error[phase_path].copy()

    def refine(self, refine_results, iter_number):
        """
        Refine the grid during subsequent iterations.

        Compute the order, number of nodes, and segment ends required for the new grid
        and assigns them to the transcription of each phase. Method of refinement is
        different for the first iteration and is done separately.

        Parameters
        ----------
        refine_results : dict
            A dictionary where each key is the path to a phase in the problem, and the
            associated value are various properties of that phase needed by the refinement
            algorithm.  refine_results is returned by check_error.  This method modifies it
            in place, adding the new_num_segments, new_order, and new_segment_ends.
        iter_number : int
            Current iteration of the grid refinement. The first iteration is 1.

        Returns
        -------
        dict
            A dictionary of phase paths : phases which were refined.
        """
        if iter_number == 1:
            self.refine_first_iter(refine_results)
            return

        x_dd = {}
        for phase_path, phase_refinement_results in refine_results.items():
            phase = self.phases[phase_path]
            self.error[phase_path] = refine_results[phase_path]['max_rel_error']
            refine_tol = phase.refine_options['tolerance']
            refine_min_order = phase.refine_options['min_order']
            tx = phase.options['transcription']
            gd = tx.grid_data

            num_scalar_states = 0
            for state_name, options in phase.state_options.items():
                shape = options['shape']
                size = np.prod(shape)
                num_scalar_states += size

            seg_order = gd.transcription_order
            seg_ends = gd.segment_ends
            numseg = gd.num_segments

            need_refine = phase_refinement_results['need_refinement']
            if not phase.refine_options['refine'] or not np.any(need_refine):
                refine_results[phase_path]['new_order'] = seg_order
                refine_results[phase_path]['new_num_segments'] = gd.num_segments
                refine_results[phase_path]['new_segment_ends'] = seg_ends
                continue

            left_end_idxs = gd.subset_node_indices['segment_ends'][0::2]
            left_end_idxs = np.append(left_end_idxs, gd.subset_num_nodes['all'])
            refine_seg_idxs = np.where(need_refine)[0]

            old_left_end_idxs = self.previous_gd[phase_path].subset_node_indices['segment_ends'][0::2]
            old_left_end_idxs = np.append(old_left_end_idxs, self.previous_gd[phase_path].subset_num_nodes['all'])

            # compute curvature
            L, D = interpolation_lagrange_matrix(gd, gd)
            x, _, _, x_d = eval_ode_on_grid(phase=phase, transcription=tx)
            x_dd[phase_path] = {}
            P = {}
            P_hat = {}
            R = np.zeros(numseg)

            # Compute the maximum magnitude of the second derivative of each state
            # Find the same value at the same time on the previous solution
            # If the ratio of these two values for a given state is highest, it is stored as the curvature
            for state_name, options in phase.state_options.items():
                x_dd[phase_path][state_name] = D @ x_d[state_name]
                P[state_name] = np.zeros(numseg)
                P_hat[state_name] = np.zeros(numseg)
                for k in np.nditer(refine_seg_idxs):
                    interp = LagrangeBarycentricInterpolant(
                        self.previous_gd[phase_path].node_stau[old_left_end_idxs[self.parent_seg_map[phase_path][k]]:
                                                               old_left_end_idxs[self.parent_seg_map[phase_path][k] +
                                                                                 1]],
                        options['shape'])
                    interp.setup(x0=-1, xf=1,
                                 f_j=self.previous_x_dd[phase_path][state_name][
                                     old_left_end_idxs[self.parent_seg_map[phase_path][k]]:old_left_end_idxs[
                                         self.parent_seg_map[phase_path][k] + 1]])
                    P[state_name][k] = np.max(
                        np.fabs(x_dd[phase_path][state_name][left_end_idxs[k]:left_end_idxs[k + 1]]))
                    xdd_max_time = gd.node_stau[left_end_idxs[k] + np.argmax(np.amax(
                        np.fabs(x_dd[phase_path][state_name][left_end_idxs[k]:left_end_idxs[k + 1]]), axis=1), axis=0)]
                    P_hat[state_name][k] = np.amax(np.fabs(interp.eval(xdd_max_time)))
                    if P[state_name][k] / P_hat[state_name][k] > R[k]:
                        R[k] = P[state_name][k] / P_hat[state_name][k]

            non_smooth_idxs = np.where(R > phase.refine_options['smoothness_factor'])[0]
            smooth_need_refine_idxs = np.setdiff1d(refine_seg_idxs, non_smooth_idxs)

            mul_factor = np.ones(numseg)
            h = 0.5 * (seg_ends[1:] - seg_ends[:-1])
            H = np.ones(numseg, dtype=int)
            h_prev = 0.5 * (self.previous_gd[phase_path].segment_ends[1:] - self.previous_gd[phase_path].segment_ends[:-1])

            split_parent_seg_idxs = self.parent_seg_map[phase_path][smooth_need_refine_idxs]

            q_smooth = (np.log(self.error[phase_path][smooth_need_refine_idxs] /
                               self.previous_error[phase_path][split_parent_seg_idxs]) +
                        2.5 * np.log(seg_order[smooth_need_refine_idxs] /
                                     self.previous_gd[phase_path].transcription_order[split_parent_seg_idxs])
                        ) / (np.log((h[smooth_need_refine_idxs] / h_prev[split_parent_seg_idxs])) +
                             np.log(seg_order[smooth_need_refine_idxs] /
                                    self.previous_gd[phase_path].transcription_order[split_parent_seg_idxs]))

            q_smooth[q_smooth < 3] = 3.0
            q_smooth[np.isposinf(q_smooth)] = 3.0
            mul_factor[smooth_need_refine_idxs] = (self.error[phase_path][smooth_need_refine_idxs] /
                                                   phase.refine_options['tolerance']) ** \
                                                  (1 / (q_smooth - 2.5))

            new_order = np.ceil(gd.transcription_order * mul_factor).astype(int)
            if gd.transcription == 'gauss-lobatto':
                odd_idxs = np.where(new_order % 2 != 0)
                new_order[odd_idxs] += 1

            split_seg_idxs = np.concatenate([np.where(new_order > phase.refine_options['max_order'])[0],
                                             non_smooth_idxs])

            check_comb_indx = np.where(np.logical_and(np.logical_and(np.logical_and(np.invert(need_refine[:-1]),
                                                                                    np.invert(need_refine[1:])),
                                                                     new_order[:-1] == new_order[1:]),
                                                      new_order[:-1] == phase.refine_options['min_order']))[0]

            reduce_order_indx = np.setdiff1d(np.where(np.invert(need_refine)), check_comb_indx)

            new_order[split_seg_idxs] = seg_order[split_seg_idxs]
            split_parent_seg_idxs = self.parent_seg_map[phase_path][split_seg_idxs]

            q_split = np.log((self.error[phase_path][split_seg_idxs] /
                              self.previous_error[phase_path][split_parent_seg_idxs]) /
                             (seg_order[split_seg_idxs] / self.previous_gd[phase_path].transcription_order[
                                 split_parent_seg_idxs]) ** 2.5
                             ) / np.log((h[split_seg_idxs] / h_prev[split_parent_seg_idxs]) /
                                        (seg_order[split_seg_idxs] / self.previous_gd[phase_path].transcription_order[
                                            split_parent_seg_idxs]))

            q_split[q_split < 3] = 3
            q_split[np.isposinf(q_split)] = 3

            H[split_seg_idxs] = np.maximum(np.minimum(
                np.ceil((self.error[phase_path][split_seg_idxs] / phase.refine_options['tolerance']
                         ) ** (1 / q_split)), np.ceil(np.log(self.error[phase_path][split_seg_idxs] /
                                                             phase.refine_options['tolerance']) /
                                                      np.log(seg_order[split_seg_idxs]))),
                2*np.ones(split_seg_idxs.size))

            # reduce segment order where error is much below the tolerance
            if reduce_order_indx.size > 0:
                # compute normalization factor beta
                beta = {}
                for state_name, options in phase.state_options.items():
                    beta[state_name] = 0
                    for k in range(0, numseg):
                        beta_seg = np.max(np.abs(x[state_name][left_end_idxs[k]:left_end_idxs[k + 1]]))
                        if beta_seg > beta[state_name]:
                            beta[state_name] = beta_seg
                    beta[state_name] += 1

                for k in np.nditer(reduce_order_indx):
                    seg_size = {'radau-ps': seg_order[k] + 1, 'gauss-lobatto': seg_order[k]}
                    if seg_order[k] == phase.refine_options['min_order']:
                        continue
                    new_order_state = {}
                    new_order[k] = seg_order[k]
                    a = np.zeros((seg_size[gd.transcription], seg_size[gd.transcription]))
                    s, _ = lgr(seg_order[k], include_endpoint=True)
                    if gd.transcription == 'gauss-lobatto':
                        s, _ = lgl(seg_order[k])
                    for j in range(0, seg_size[gd.transcription]):
                        roots = s[s != s[j]]
                        Q = np.poly(roots)
                        a[:, j] = Q / np.polyval(Q, s[j])

                    for state_name, options in phase.state_options.items():
                        new_order_state[state_name] = seg_order[k]
                        b = a @ x[state_name][left_end_idxs[k]:left_end_idxs[k + 1]]
                        for i in range(seg_size[gd.transcription] - 1, phase.refine_options['min_order'], -1):
                            state_error = np.abs(b[i]) / beta[state_name]
                            if state_error < refine_tol:
                                if gd.transcription == 'gauss-lobatto':
                                    if new_order_state[state_name] < i - 2 <= refine_min_order:
                                        new_order_state[state_name] = i - 2
                                elif gd.transcription == 'radau-ps':
                                    if new_order_state[state_name] < i - 1 <= refine_min_order:
                                        new_order_state[state_name] = i - 1
                        new_order[k] = max(new_order_state.values())

            # combine unnecessary segments
            merge_seg = np.zeros(numseg, dtype=bool)
            if check_comb_indx.size > 0:
                for k in np.nditer(check_comb_indx):
                    seg_size = {'radau-ps': seg_order[k] + 1, 'gauss-lobatto': seg_order[k]}
                    if merge_seg[k]:
                        continue

                    a = np.zeros((seg_size[gd.transcription], seg_size[gd.transcription]))
                    h_ = np.maximum(h[k], h[k + 1])
                    s, _ = lgr(new_order[k].astype(int), include_endpoint=True)
                    if gd.transcription == 'gauss-lobatto':
                        s, _ = lgl(new_order[k])
                    for j in range(0, seg_size[gd.transcription]):
                        roots = s[s != s[j]]
                        Q = np.poly(roots)
                        a[:, j] = Q / np.polyval(Q, s[j])

                    merge_seg[k + 1] = True

                    for state_name, options in phase.state_options.items():
                        beta = 1 + np.max(x[state_name][left_end_idxs[k]:left_end_idxs[k + 1]])
                        c = a @ x[state_name][left_end_idxs[k]:left_end_idxs[k + 1]]
                        b = np.multiply(c.ravel(), np.array([(h_ / h[k]) ** l for l in
                                                             range(seg_size[gd.transcription])]))
                        b_hat = np.multiply(c.ravel(),
                                            np.array([(h_ / h[k + 1]) ** l for l in range(seg_size[gd.transcription])]))
                        err_val = np.dot(np.absolute(b - b_hat).ravel(),
                                         np.array([2 ** l for l in range(seg_size[gd.transcription])])) / beta

                        if err_val > phase.refine_options['tolerance']/10 and merge_seg[k + 1]:
                            merge_seg[k + 1] = False

            H[np.where(merge_seg)] = 0

            new_order = np.repeat(new_order, repeats=H)
            new_num_segments = int(np.sum(H))
            new_segment_ends = split_segments(gd.segment_ends, H)

            if gd.transcription == 'gauss-lobatto':
                new_order[new_order % 2 == 0] = new_order[new_order % 2 == 0] + 1

            self.parent_seg_map[phase_path] = np.zeros(new_num_segments, dtype=int)
            for i in range(1, new_num_segments):
                for j in range(1, numseg):
                    if new_segment_ends[i] == gd.segment_ends[j]:
                        self.parent_seg_map[phase_path][i] = int(j)
                        break
                    elif gd.segment_ends[j - 1] < new_segment_ends[i] < gd.segment_ends[j]:
                        self.parent_seg_map[phase_path][i] = int(j - 1)
                        break

            refine_results[phase_path]['new_order'] = new_order
            refine_results[phase_path]['new_num_segments'] = new_num_segments
            refine_results[phase_path]['new_segment_ends'] = new_segment_ends

            tx.options['order'] = new_order
            tx.options['num_segments'] = new_num_segments
            tx.options['segment_ends'] = new_segment_ends
            tx.init_grid()
            self.previous_x_dd[phase_path] = x_dd[phase_path].copy()
            self.previous_error[phase_path] = self.error[phase_path].copy()
            self.previous_gd[phase_path] = copy.deepcopy(gd)
